#!/usr/bin/env python3

CREATE      =b'\x00' # sphinx
READ        =b'\x33' # blob
UNDO        =b'\x55' # change sphinx
GET         =b'\x66' # sphinx
V1GET       =b'\x69' # v1 sphinx
COMMIT      =b'\x99' # change sphinx
CHANGE_DKG  =b'\xa0' # sphinx/dkg
CHANGE      =b'\xaa' # sphinx
CREATE_DKG  =b'\xf0' # sphinx/dkg
V1DELETE    =b'\xf9' # v1 sphinx+blobs
DELETE      =b'\xff' # sphinx+blobs

CHALLENGE_CREATE = b'\x5a'
CHALLENGE_VERIFY = b'\xa5'

VERSION = b'\x01'

V1RULE_SIZE = 79
RULE_SIZE = V1RULE_SIZE+32



#!/usr/bin/env python

import struct, json, sys, subprocess

if sys.argv[1]=="get":
    msg = {'cmd': "login"}
else:
    msg = {'cmd': sys.argv[1]}

if sys.argv[1] in ('create', 'get', 'change', 'commit', 'undo', 'delete'):
    msg['name']= sys.argv[2]
    msg['site']= sys.argv[3]

if sys.argv[1] in {'create', 'change'}:
    msg['rules']= sys.argv[4]
    msg['size']= sys.argv[5]

if sys.argv[1] == 'list':
    msg['site']= sys.argv[2]

msg['mode'] = 'ws-test'

if sys.argv[1] == 'json':
    msg = sys.argv[2].replace("'", '"')
else:
    msg = json.dumps(msg)

print("cmd:", msg)
cmd = struct.pack('i', len(msg))+msg.encode("utf-8")
proc=subprocess.Popen(["../pwdsphinx/websphinx.py"], stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
out, err = proc.communicate(input=cmd)
print("ret", proc.returncode)
print("stdout")
for line in out.split(b'\n'):
    print(line)
print("stderr")
for line in err.split(b'\n'):
    print(line)



import unittest
from os import listdir, makedirs, environ, path
from shutil import rmtree, copyfile
from tempfile import mkdtemp
from unittest.mock import Mock
from io import BytesIO, StringIO
import sys, pysodium, subprocess, time, struct
import tracemalloc
from pyoprf import multiplexer
from pwdsphinx import sphinx, bin2pass, ostore, v1sphinx
from binascii import b2a_base64, a2b_base64
import pyoprf, ctypes
import contextlib

#don't run with -b if you want log output from setup/teardown fns
#import logging
#logger = logging.getLogger(__name__)

# to get coverage, run
# PYTHONPATH=.. coverage run ../tests/test.py
# coverage report -m
# to just run the tests do
# python3 -m unittest discover --start-directory ../tests

# disable the output of sphinx
sphinx.print = Mock()

N = 3
data_dir = 'data/'
char_classes = 'uld'
syms = bin2pass.symbols
size = 0
pwd = 'asdf'
user = 'user1'
user2 = 'user2'
host = 'example.com'
servers = {'zero': {'host': 'localhost', 'port': 10000, 'ssl_cert': 'cert.pem'},
           'one':  {'host': 'localhost', 'port': 10001, 'ssl_cert': 'cert.pem'},
           'two':  {'host': 'localhost', 'port': 10002, 'ssl_cert': 'cert.pem'},
           'drei': {'host': 'localhost', 'port': 10003, 'ssl_cert': 'cert.pem'},
           'eris': {'host': 'localhost', 'port': 10004, 'ssl_cert': 'cert.pem'}
           }
corrupt_dkg_lib = environ.get('CORRUPT_DKG_LIB')
ostore_server = environ.get('OPAQUESTORE_SERVER')
orig_servers=sphinx.servers
max_recovery_tokens = 2
ostore_max_fails = 3

import random
random.seed()
unittest.TestLoader.sortTestMethodsUsing = lambda _, x, y: random.randint(-1,1)

class Input:
  def __init__(self, txt = None, pwd = pwd):
    if txt:
      self.buffer = BytesIO('\n'.join((pwd, txt)).encode())
    else:
      self.buffer = BytesIO(pwd.encode())
  def isatty(self):
      return False
  def close(self):
    return

def connect(peers=None):
  if peers == None:
    peers = dict(tuple(servers.items())[:N])
  #m = multiplexer.Multiplexer(peers)
  m = sphinx.Multiplexer(peers)
  m.connect()
  return m

def bad_signkey(_, __):
  pk, sk = pysodium.crypto_sign_seed_keypair(b'\xfe'*pysodium.crypto_sign_SEEDBYTES)
  return sk, pk
get_signkey = sphinx.get_signkey

class TestEndToEnd(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
      #libc = ctypes.cdll.LoadLibrary('libc.so.6')
      #cstderr = ctypes.c_void_p.in_dll(libc, 'stderr')
      #log_file = ctypes.c_void_p.in_dll(pyoprf.liboprf,'log_file')
      #log_file.value = cstderr.value

      cls._validate_password = sphinx.validate_password

      cls._root = mkdtemp(prefix='sphinx-oracle-root.')
      root = cls._root
      pks = []
      for idx, k in enumerate(servers.keys()):
        makedirs(f"{root}/servers/{idx}")
        copyfile("cert.pem", f"{root}/servers/{idx}/cert.pem")
        copyfile("key.pem", f"{root}/servers/{idx}/key.pem")
        pk, sk = pysodium.crypto_sign_keypair()
        with open(f"{root}/servers/{idx}/ltsig.key", 'wb') as fd:
          fd.write(sk)
        servers[k]['ltsigkey']=pk
        if k in sphinx.servers:
          sphinx.servers[k]['ltsigkey']=pk
          sphinx.servers[k]['timeout']=30
        # for reproducability, in case we want to run manual sphinx ops
        with open(f"data/{k}.pub",'wb') as fd:
          fd.write(pk)

        with open(f"{root}/servers/{idx}/sphinx.cfg", 'w') as fd:
          fd.write(f'[server]\n'
                   f'verbose = true\n'
                   f'address = "127.0.0.1"\n'
                   f'port={10000+idx}\n'
                   f'timeout = 30\n'
                   f'max_kids = 5\n'
                   f'ssl_key= "key.pem"\n'
                   f'ssl_cert= "cert.pem"\n'
                   f'ltsigkey = "ltsig.key"\n'
                   f'datadir = "data"\n'
                   f'rl_decay=1800\n'
                   f'rl_threshold=10\n')
      cls._oracles = []
      env = environ
      for idx in range(len(servers)):
        log = open(f"{root}/servers/{idx}/log", "w")
        if idx == N and corrupt_dkg_lib is not None:
          print(f"enabling byzantine peers {corrupt_dkg_lib}", file=log)
          env["BYZANTINE_DKG"]=corrupt_dkg_lib
        cls._oracles.append(
          (subprocess.Popen(["python3", path.dirname(path.abspath(sphinx.__file__)) + "/oracle.py"], cwd = f"{root}/servers/{idx}/", stdout=log, stderr=log, pass_fds=[log.fileno()], env=env), log))
        log.close()
      if corrupt_dkg_lib is not None:
        del env["BYZANTINE_DKG"]

      if ostore.available and ostore_server is not None and path.isfile(ostore_server):
        cls._ostore_root = mkdtemp(prefix='opaquestore-server-root.')
        root = cls._ostore_root
        pks = []
        for idx in range(len(servers)):
          makedirs(f"{root}/servers/{idx}")
          copyfile("cert.pem", f"{root}/servers/{idx}/cert.pem")
          copyfile("key.pem", f"{root}/servers/{idx}/key.pem")
          pk, sk = pysodium.crypto_sign_keypair()
          with open(f"{root}/servers/{idx}/ltsig.key", 'wb') as fd:
            fd.write(sk)
          #pks.append(b2a_base64(pk).decode("utf8")[:-1])
          pks.append(pk)
          with open(f"{root}/servers/{idx}/opaque-stored.cfg", 'w') as fd:
            fd.write(f'[server]\n'
                     f'verbose = true\n'
                     f'address = "127.0.0.1"\n'
                     f'port={23000+idx}\n'
                     f'timeout = 30\n'
                     f'max_kids = 5\n'
                     f'ssl_key= "key.pem"\n'
                     f'ssl_cert= "cert.pem"\n'
                     f'ltsigkey = "ltsig.key"\n'
                     f'record_salt = "some random string to salt the record ids"\n'
                     f'max_blob_size = 8192\n'
                     f'max_recovery_tokens = {max_recovery_tokens}\n'
                     f'max_fails = {ostore_max_fails}\n'
                     f'datadir = "data"\n\n')
        # lt sig pubkeys
        for idx in range(len(servers)):
          for pk, name in zip(pks, servers.keys()):
            with open(f"data/os_{name}.pub",'wb') as fd:
              fd.write(pk)
        for idx in range(len(servers)):
          log = open(f"{root}/servers/{idx}/log", "w")
          cls._oracles.append(
            (subprocess.Popen([ostore_server], cwd = f"{root}/servers/{idx}/", stdout=log, stderr=log, pass_fds=[log.fileno()]), log))
          log.close()
      time.sleep(0.8)

    @classmethod
    def tearDownClass(cls):
      for p, log in cls._oracles:
        p.kill()
        r = p.wait()
        log.close()
      #rmtree(cls._root)
      time.sleep(0.4)

    def tearDown(self):
      sphinx.validate_password=self._validate_password
      time.sleep(3)
      sphinx.servers = orig_servers
      #cleanup()
      roots = [self._root]
      if hasattr(self, '_ostore_root'):
        roots.append(self._ostore_root)
      for idx in range(len(servers)):
        for root in roots:
          ddir = f"{root}/servers/{idx}/data/"
          if not path.exists(ddir): continue
          for f in listdir(ddir):
            if f == 'key': continue
            rmtree(ddir+f)

    def test_create_user(self):
        with connect() as s:
            self.assertIsInstance(sphinx.create(s, pwd, user, host, char_classes, syms, size), str)

    def test_huge_user(self):
        if sphinx.userlist == False: return
        with connect() as s:
            self.assertRaises(ValueError, sphinx.create,s, pwd, 'a'*(2**16 - 40), host, char_classes, syms, size)
        with connect() as s:
            rwd=sphinx.create(s, pwd, 'a'*(2**16 - 42), host, char_classes, syms, size)
            self.assertIsInstance(rwd, str)
        with connect() as s:
            self.assertRaises(ValueError, sphinx.create, s, pwd, 'a', host, char_classes, syms, size)

    def test_rules_u(self):
        with connect() as s:
            rwd = sphinx.create(s, pwd, user, host, "u", '', 0)
        self.assertIsInstance(rwd, str)
        self.assertTrue(rwd.isupper())

    def test_rules_l(self):
        with connect() as s:
            rwd = sphinx.create(s, pwd, user, host, "l", '', 0)
        self.assertIsInstance(rwd, str)
        self.assertTrue(rwd.islower())

    def test_rules_d(self):
        with connect() as s:
            rwd = sphinx.create(s, pwd, user, host, "d", '', 0)
        self.assertIsInstance(rwd, str)
        self.assertTrue(rwd.isdigit())

    def test_rules_ulsd(self):
        with connect() as s:
            rwd = sphinx.create(s, pwd, user, host, char_classes, syms, 0)
        self.assertIsInstance(rwd, str)
        self.assertTrue(len(set([x.decode('utf8') for x in bin2pass.sets['u']]).intersection(rwd)) > 0)
        self.assertTrue(len(set([x.decode('utf8') for x in bin2pass.sets['l']]).intersection(rwd)) > 0)
        self.assertTrue(len(set([x.decode('utf8') for x in bin2pass.sets['d']]).intersection(rwd)) > 0)
        self.assertTrue(len(set(bin2pass.symbols).intersection(rwd)) > 0)

    def test_pwd_len(self):
        for i in range(1,32):
            with connect() as s:
                rwd = sphinx.create(s, pwd, user, host, char_classes, syms, i)
                self.assertIsInstance(rwd, str)
                self.assertTrue(len(rwd)==i)
            with connect() as s:
                self.assertTrue(sphinx.delete(s, pwd, user, host))

    def test_invalid_rules(self):
        with connect() as s:
            self.assertRaises(ValueError, sphinx.create, s, pwd, user, host, "asdf", syms, size)

    def test_recreate_user(self):
        with connect() as s:
            self.assertIsInstance(sphinx.create(s, pwd, user, host, char_classes, syms, size), str)

        with connect() as s:
            self.assertRaises(ValueError, sphinx.create,s, pwd, user, host, char_classes, syms, size)
            s.close()

    def test_get_nonexistant_record(self):
        for i in [0, 1,2]:
          try: makedirs(f"{self._root}/servers/{i}/data/")
          except: pass
        with connect() as s:
            self.assertRaises(ValueError, sphinx.get, s, pwd, user, host)

    def test_get(self):
        with connect() as s:
            rwd0 = sphinx.create(s, pwd, user, host, char_classes, syms, size)
            self.assertIsInstance(rwd0, str)

        with connect() as s:
            rwd = sphinx.get(s, pwd, user, host)
        self.assertIsInstance(rwd, str)

        self.assertEqual(rwd,rwd0)

    def test_v1get(self):
        if not v1sphinx.enabled: return
        # synthetically create a v1 record
        id = v1sphinx.getid(host,user)
        for i in [1,2]:
          try: makedirs(f"{self._root}/servers/{i}/data/")
          except: pass
        ddir = f"{self._root}/servers/0/data/{id.hex()}"
        k = b'\x55' * 32
        #calculate rwd
        h0 = pysodium.crypto_generichash(pwd.encode(), outlen=pysodium.crypto_core_ristretto255_HASHBYTES);
        H0 = pysodium.crypto_core_ristretto255_from_hash(h0)
        H0_k = pysodium.crypto_scalarmult_ristretto255(k, H0)
        rwd0 = pysodium.crypto_generichash(pwd.encode()+H0_k, outlen=pysodium.crypto_core_ristretto255_BYTES);
        rwd0 = pysodium.crypto_pwhash(pysodium.crypto_core_ristretto255_BYTES,
                                      rwd0, id[:pysodium.crypto_pwhash_SALTBYTES],
                                      pysodium.crypto_pwhash_OPSLIMIT_INTERACTIVE,
                                      pysodium.crypto_pwhash_MEMLIMIT_INTERACTIVE)
        # create a rules blob
        if sphinx.validate_password:
            check_digit = pysodium.crypto_generichash(v1sphinx.CHECK_CTX, rwd0, 1)[0]
        else:
            check_digit = 0
        xor_mask = b'\x55' * 32
        packed = 30
        packed = packed + (sum(1<<i for i, c in enumerate(('u','l','d'))) << 7)
        packed = packed + (sum(1<<i for i, c in enumerate(bin2pass.symbols)) << (7 + 3))
        packed = packed + ((check_digit & ((1<<5) - 1)) << (7 + 3 + 33) )
        pt = packed.to_bytes(6,"big") + xor_mask
        rules = v1sphinx.encrypt_blob(pt)

        makedirs(ddir)
        with open(ddir+"/key", 'wb') as fd:
            fd.write(k)
        with open(ddir+"/pub", 'wb') as fd:
            sk, pk = v1sphinx.get_signkey(id, rwd0)
            fd.write(pk)
        with open(ddir+"/rules", 'wb') as fd:
            fd.write(rules)

        # create also a v1 users blob
        blobid=v1sphinx.getid(host,'')
        bddir = f"{self._root}/servers/0/data/{blobid.hex()}"
        makedirs(bddir)
        blob = user.encode()
        blob = v1sphinx.encrypt_blob(blob)
        blob = struct.pack("!H", len(blob)) + blob
        with open(bddir+"/pub", 'wb') as fd:
            sk, pk = v1sphinx.get_signkey(blobid, b'')
            fd.write(pk)
        with open(bddir+"/blob", 'wb') as fd:
            fd.write(blob)

        with connect() as s:
            rwd = sphinx.get(s, pwd, user, host)

        self.assertIsInstance(rwd, str)
        self.assertEqual(rwd,'_HO; <Yk)KA:G.q@8\\6zVHtDttCRA\\')

        self.assertTrue(not path.exists(ddir))
        self.assertTrue(not path.exists(bddir))
        # try to get the value now from the uplifted v2 record
        with connect() as s:
            rwd1 = sphinx.get(s, pwd, user, host)

        self.assertEqual(rwd,rwd1)

    def test_v1users(self):
        if not v1sphinx.enabled: return
        if sphinx.userlist == False: return

        with connect() as s:
            self.assertIsInstance(sphinx.create(s, pwd, user, host, char_classes, syms, size), str)

        with connect() as s:
            self.assertIsInstance(sphinx.create(s, pwd, user2, host, char_classes, syms, size), str)

        with connect() as s:
            users = sphinx.users(s, host)
            self.assertIsInstance(users, str)
            self.assertEqual(users, '\n'.join((user,user2)))

        # synthetically create a v1 record
        id = v1sphinx.getid(host,'')
        for i in [1,2]:
          try: makedirs(f"{self._root}/servers/{i}/data/")
          except: pass
        ddir = f"{self._root}/servers/0/data/{id.hex()}"
        makedirs(ddir)

        v1users = {'v1user1', 'v1user2', 'v1used'}
        blob = ('\x00'.join(sorted(v1users))).encode()
        # notice we do not add rwd to encryption of user blobs
        blob = v1sphinx.encrypt_blob(blob)
        bsize = len(blob)
        blob = struct.pack("!H", bsize) + blob
        blob = v1sphinx.sign_blob(blob, id, b'')

        with open(ddir+"/pub", 'wb') as fd:
            sk, pk = v1sphinx.get_signkey(id, b'')
            fd.write(pk)
        with open(ddir+"/blob", 'wb') as fd:
            fd.write(blob)

        with connect() as s:
            users = sphinx.users(s, host)
            self.assertIsInstance(users, str)
            users = set(users.split('\n'))
            self.assertEqual(users, {'user1', 'user2'} | v1users)

    #def test_get_inv_mpwd(self):
    #    if not sphinx.validate_password:
    #        return
    #    with connect() as s:
    #        rwd0 = sphinx.create(s, pwd, user, host, char_classes, syms, size)
    #        self.assertIsInstance(rwd0, str)

    #    with connect() as s:
    #        self.assertRaises(ValueError, sphinx.get, s, 'zxcv1', user, host)

    def test_get_nonexistant_host(self):
        with connect() as s:
            self.assertRaises(ValueError, sphinx.get, s, pwd, user, host)

    def test_delete(self):
        with connect() as s:
            self.assertIsInstance(sphinx.create(s, pwd, user, host, char_classes, syms, size), str)

        with connect() as s:
            self.assertTrue(sphinx.delete(s, pwd, user, host))

    def test_delete_inv_mpwd(self):
        if sphinx.rwd_keys == False: return
        with connect() as s:
            self.assertIsInstance(sphinx.create(s, pwd, user, host, char_classes, syms, size), str)

        with connect() as s:
            self.assertRaises(ValueError, sphinx.delete, s, 'zxcv', user, host)

    def test_change(self):
        with connect() as s:
            self.assertIsInstance(sphinx.create(s, pwd, user, host, char_classes, syms, size), str)

        with connect() as s:
            pwd0 = sphinx.get(s, pwd, user, host)
        self.assertIsInstance(pwd0, str)

        with connect() as s:
            pwd1 = sphinx.change(s, pwd, pwd, user, host)
        self.assertIsInstance(pwd1, str)
        self.assertNotEqual(pwd0, pwd1)

        with connect() as s:
            pwd2 = sphinx.change(s, pwd, pwd.upper(), user, host)
        self.assertIsInstance(pwd2, str)
        self.assertNotEqual(pwd0, pwd2)
        self.assertNotEqual(pwd1, pwd2)

    def test_commit_undo(self):
        # create
        with connect() as s:
            pwd0 = sphinx.create(s, pwd, user, host, char_classes, syms, size)
        self.assertIsInstance(pwd0, str)

        # get
        with connect() as s:
            pwd1 = sphinx.get(s, pwd, user, host)
        self.assertIsInstance(pwd1, str)
        self.assertEqual(pwd0, pwd1)

        # change
        with connect() as s:
            pwd2 = sphinx.change(s, pwd, pwd.upper(), user, host)
        self.assertIsInstance(pwd2, str)
        self.assertNotEqual(pwd1, pwd2)

        # get
        with connect() as s:
            pwd3 = sphinx.get(s, pwd, user, host)
        self.assertIsInstance(pwd3, str)
        self.assertEqual(pwd1, pwd3)

        # commit
        with connect() as s:
            sphinx.commit(s, pwd, user, host)
        with connect() as s:
            pwd4 = sphinx.get(s, pwd.upper(), user, host)
        self.assertIsInstance(pwd4, str)
        self.assertEqual(pwd2, pwd4)

        # undo
        with connect() as s:
            sphinx.undo(s, pwd.upper(), user, host, )
        with connect() as s:
            pwd5 = sphinx.get(s, pwd, user, host)
        self.assertIsInstance(pwd5, str)
        self.assertEqual(pwd1, pwd5)

    def test_commit_undo_inv_mpwd(self):
        # create
        if sphinx.rwd_keys == False: return
        with connect() as s:
            pwd0 = sphinx.create(s, pwd, user, host, char_classes, syms, size)
            self.assertIsInstance(pwd0, str)

        # change invalid mpwd
        with connect() as s:
           self.assertRaises(ValueError, sphinx.change,s, 'zxcv', pwd, user, host)

        # change correct mpwd
        with connect() as s:
           pwd2 = sphinx.change(s, pwd, pwd, user, host)
        self.assertIsInstance(pwd2, str)
        self.assertNotEqual(pwd0, pwd2)

        # commit invalid mpwd
        with connect() as s:
           self.assertRaises(ValueError, sphinx.commit,s, 'zxcv', user, host)

        # commit correct mpwd
        with connect() as s:
           sphinx.commit(s, pwd, user, host)
        with connect() as s:
           pwd4 = sphinx.get(s, pwd, user, host)
        self.assertIsInstance(pwd4, str)
        self.assertEqual(pwd2, pwd4)

        # undo invalid mpwd
        with connect() as s:
           self.assertRaises(ValueError, sphinx.undo,s, 'zxcv', user, host)

        # undo correct mpwd
        with connect() as s:
           sphinx.undo(s, pwd, user, host)
        with connect() as s:
           pwd5 = sphinx.get(s, pwd, user, host)
        self.assertIsInstance(pwd5, str)
        self.assertEqual(pwd0, pwd5)

    def test_list_users(self):
        if sphinx.userlist == False: return
        with connect() as s:
            self.assertIsInstance(sphinx.create(s, pwd, user, host, char_classes, syms, size), str)
        with connect() as s:
            self.assertIsInstance(sphinx.create(s, pwd, user2, host, char_classes, syms, size), str)
        with connect() as s:
            users = sphinx.users(s, host)
            self.assertIsInstance(users, str)
            self.assertEqual(users, '\n'.join((user,user2)))

    def test_list_users_diff_mpwd(self):
        if sphinx.userlist == False: return
        with connect() as s:
            self.assertIsInstance(sphinx.create(s, pwd, user, host, char_classes, syms, size), str)
        with connect() as s:
            self.assertIsInstance(sphinx.create(s, 'zxcv', user2, host, char_classes, syms, size), str)
        with connect() as s:
            users = sphinx.users(s, host)
            self.assertIsInstance(users, str)
            self.assertEqual(users, '\n'.join((user,user2)))

    def test_double_commit(self):
        # create
        with connect() as s:
            pwd0 = sphinx.create(s, pwd, user, host, char_classes, syms, size)
            self.assertIsInstance(pwd0, str)

        # change
        with connect() as s:
            pwd2 = sphinx.change(s, pwd, pwd, user, host)
        self.assertIsInstance(pwd2, str)
        self.assertNotEqual(pwd0, pwd2)

        # commit
        with connect() as s:
            sphinx.commit(s, pwd, user, host)
        with connect() as s:
            pwd4 = sphinx.get(s, pwd, user, host)
        self.assertIsInstance(pwd4, str)
        self.assertEqual(pwd2, pwd4)

        # commit
        with connect() as s:
            self.assertRaises(ValueError, sphinx.commit,s, pwd, user, host)

    def test_auth(self):
        # create
        with connect() as s:
            rwd = sphinx.create(s, pwd, user, host, char_classes, syms, size)
            self.assertIsInstance(rwd, str)
        sphinx.get_signkey = bad_signkey
        with connect() as s:
             self.assertRaises(ValueError, sphinx.change, s, pwd, pwd, user, host, char_classes, syms, size)
        sphinx.get_signkey = get_signkey

    def test_userblob_auth_create(self):
        if sphinx.userlist == False: return
        # create
        with connect() as s:
            rwd = sphinx.create(s, pwd, user, host, char_classes, syms, size)
            self.assertIsInstance(rwd, str)
        sphinx.get_signkey = bad_signkey
        with connect() as s:
            self.assertRaises(ValueError, sphinx.create, s, pwd, user2, host, char_classes, syms, size)
        sphinx.get_signkey = get_signkey

    def test_create_user_xormask(self):
        with connect() as s:
          rwd = sphinx.create(s, pwd, user, host, '', '', 0, pwd)
        self.assertIsInstance(rwd, str)
        self.assertEqual(pwd, rwd)

    def test_change_xormask(self):
        with connect() as s:
          rwd0 = sphinx.create(s, pwd, user, host, char_classes, syms, size)
          self.assertIsInstance(rwd0, str)

        with connect() as s:
            rwd1 = sphinx.change(s, pwd, pwd, user, host, '', '', 0, pwd)
        self.assertIsInstance(rwd1, str)
        self.assertEqual(rwd1, pwd)

        with connect() as s:
            rwd2 = sphinx.change(s, pwd, pwd, user, host, '', '', 0, pwd+pwd)
        self.assertIsInstance(rwd2, str)
        self.assertEqual(rwd2, pwd+pwd)

    def test_corrupted_dkg(self):
        sphinx.servers = {
          'zero': servers['zero'],
          'drei': servers['drei'],
          'eris': servers['eris']
        }

        if isinstance(self, TestEndToEndSingleMode):
          return
        if corrupt_dkg_lib is None:
          # skipping since we don't have the byzantine peers lib
          return
        with connect(sphinx.servers) as s:
          self.assertRaises(ValueError, sphinx.create ,s, pwd, user, host, char_classes, syms, size)

    def test_main_create(self):
        sys.stdin = Input()
        self.assertIsNone(sphinx.main(('sphinx.py', 'create', user, host, char_classes, syms, str(size))))

    def test_main_get(self):
        sys.stdin = Input()
        self.assertIsNone(sphinx.main(('sphinx.py', 'create', user, host, char_classes, syms, str(size))))
        sys.stdin = Input()
        self.assertIsNone(sphinx.main(('sphinx.py', 'get', user, host)))

    def test_main_delete(self):
        sys.stdin = Input()
        self.assertIsNone(sphinx.main(('sphinx.py', 'create', user, host, char_classes, syms, str(size))))
        sys.stdin = Input()
        self.assertIsNone(sphinx.main(('sphinx.py', 'delete', user, host)))

    def test_main_change_commit_undo(self):
        sys.stdin = Input("qwer")
        self.assertIsNone(sphinx.main(('sphinx.py', 'create', user, host, char_classes, syms, str(size))))
        sys.stdin = Input()
        self.assertIsNone(sphinx.main(('sphinx.py', 'change', user, host, char_classes, syms, str(size))))
        sys.stdin = Input()
        self.assertIsNone(sphinx.main(('sphinx.py', 'commit', user, host)))
        sys.stdin = Input()
        self.assertIsNone(sphinx.main(('sphinx.py', 'undo', user, host)))

    def test_main_inv_params(self):
        for cmd in ('create','get','change','commit','undo','delete','list'):
            self.assertRaises(SystemExit, sphinx.main, ('sphinx.py', cmd))

    def test_predefined_pwd(self):
        with connect() as s:
            rwd0 = sphinx.create(s, pwd, user, host, char_classes, syms, size, target = pwd)
            self.assertIsInstance(rwd0, str)
        self.assertEqual(pwd,rwd0)

        with connect() as s:
            rwd = sphinx.get(s, pwd, user, host)
        self.assertIsInstance(rwd, str)

        self.assertEqual(rwd,rwd0)

    def test_predefined_raw(self):
        target = b'A' * 32
        with connect() as s:
            rwd0 = sphinx.create(s, pwd, 'raw://'+user, host, '', '', 0, target = target)
        self.assertEqual(target,rwd0)

        with connect() as s:
            rwd = sphinx.get(s, pwd, 'raw://'+user, host)

        self.assertEqual(rwd,rwd0)

    def test_ostore_store(self):
      if not ostore.available or ostore_server is None: return
      sys.stdin = Input()
      self.assertIsNone(sphinx.main(('sphinx.py', 'store', user, 'opaque-store.cfg')))

    def test_ostore_read(self):
      if not ostore.available or ostore_server is None: return
      sys.stdin = Input()
      self.assertIsNone(sphinx.main(('sphinx.py', 'store', user, 'opaque-store.cfg')))

      sys.stdin = Input()
      f = StringIO()
      with contextlib.redirect_stdout(f):
        self.assertIsNone(sphinx.main(('sphinx.py', 'read', user)))
      with open('opaque-store.cfg','r') as fd:
        cfg = fd.read()
      self.assertTrue(cfg in f.getvalue())

    def test_ostore_read_invpwd(self):
      if not ostore.available or ostore_server is None: return
      vp = sphinx.validate_password
      sphinx.validate_password = False
      sys.stdin = Input()
      self.assertIsNone(sphinx.main(('sphinx.py', 'store', user, 'opaque-store.cfg')))

      sys.stdin = Input(pwd='qwer')
      with self.assertRaises(SystemExit) as cm:
          ret = sphinx.main(('sphinx.py', 'read', user))
          self.assertIsNone(ret)
          self.assertEqual(cm.exception.code, 1)

      sphinx.validate_password = vp

    def test_ostore_replace(self):
      if not ostore.available or ostore_server is None: return
      sys.stdin = Input()
      self.assertIsNone(sphinx.main(('sphinx.py', 'store', user, 'opaque-store.cfg')))

      sys.stdin = Input()
      f = StringIO()
      with contextlib.redirect_stdout(f):
        self.assertIsNone(sphinx.main(('sphinx.py', 'read', user)))
      with open('opaque-store.cfg','r') as fd:
        cfg = fd.read()
      self.assertTrue(cfg in f.getvalue())

      sys.stdin = Input()
      self.assertIsNone(sphinx.main(('sphinx.py', 'replace', user, 'sphinx.cfg')))

      sys.stdin = Input()
      f = StringIO()
      with contextlib.redirect_stdout(f):
        self.assertIsNone(sphinx.main(('sphinx.py', 'read', user)))
      with open('sphinx.cfg','r') as fd:
        cfg = fd.read()
      self.assertTrue(cfg in f.getvalue())

    def test_ostore_replace_invpwd(self):
      if not ostore.available or ostore_server is None: return
      sys.stdin = Input()
      self.assertIsNone(sphinx.main(('sphinx.py', 'store', user, 'opaque-store.cfg')))

      sys.stdin = Input()
      f = StringIO()
      with contextlib.redirect_stdout(f):
        self.assertIsNone(sphinx.main(('sphinx.py', 'read', user)))
      with open('opaque-store.cfg','r') as fd:
        cfg = fd.read()
      self.assertTrue(cfg in f.getvalue())

      vp = sphinx.validate_password
      sphinx.validate_password = False

      sys.stdin = Input(pwd='qwer')
      with self.assertRaises(SystemExit) as cm:
          ret = sphinx.main(('sphinx.py', 'replace', user, 'sphinx.cfg'))
          self.assertIsNone(ret)
          self.assertEqual(cm.exception.code, 1)

      sys.stdin = Input()
      f = StringIO()
      with contextlib.redirect_stdout(f):
        self.assertIsNone(sphinx.main(('sphinx.py', 'read', user)))
      with open('opaque-store.cfg','r') as fd:
        cfg = fd.read()
      self.assertTrue(cfg in f.getvalue())

    def test_ostore_erase(self):
      if not ostore.available or ostore_server is None: return
      sys.stdin = Input()
      self.assertIsNone(sphinx.main(('sphinx.py', 'store', user, 'opaque-store.cfg')))

      sys.stdin = Input()
      f = StringIO()
      with contextlib.redirect_stdout(f):
        self.assertIsNone(sphinx.main(('sphinx.py', 'read', user)))
      with open('opaque-store.cfg','r') as fd:
        cfg = fd.read()
      self.assertTrue(cfg in f.getvalue())

      sys.stdin = Input()
      self.assertIsNone(sphinx.main(('sphinx.py', 'erase', user)))

      sys.stdin = Input()
      with self.assertRaises(SystemExit) as cm:
          ret = sphinx.main(('sphinx.py', 'read', user))
          self.assertIsNone(ret)
          self.assertEqual(cm.exception.code, 1)

    def test_ostore_erase_invpwd(self):
      if not ostore.available or ostore_server is None: return
      sys.stdin = Input()
      self.assertIsNone(sphinx.main(('sphinx.py', 'store', user, 'opaque-store.cfg')))

      sys.stdin = Input()
      f = StringIO()
      with contextlib.redirect_stdout(f):
        self.assertIsNone(sphinx.main(('sphinx.py', 'read', user)))
      with open('opaque-store.cfg','r') as fd:
        cfg = fd.read()
      self.assertTrue(cfg in f.getvalue())

      vp = sphinx.validate_password
      sphinx.validate_password = False

      sys.stdin = Input(pwd='qwer')
      with self.assertRaises(SystemExit) as cm:
          ret = sphinx.main(('sphinx.py', 'erase', user))
          self.assertIsNone(ret)
          self.assertEqual(cm.exception.code, 1)

      sys.stdin = Input()
      f = StringIO()
      with contextlib.redirect_stdout(f):
        self.assertIsNone(sphinx.main(('sphinx.py', 'read', user)))
      with open('opaque-store.cfg','r') as fd:
        cfg = fd.read()
      self.assertTrue(cfg in f.getvalue())

    def test_ostore_recoverytokens(self):
      if not ostore.available or ostore_server is None: return
      sys.stdin = Input()
      self.assertIsNone(sphinx.main(('sphinx.py', 'store', user, 'opaque-store.cfg')))

      tokens = {i: set() for i in range(len(ostore.client.config['servers']))}
      for _ in range(max_recovery_tokens * 3):
        sys.stdin = Input()
        f = StringIO()
        with contextlib.redirect_stdout(f):
          self.assertIsNone(sphinx.main(('sphinx.py', 'recovery-tokens', user)))
        lines = f.getvalue().split('\n')
        self.assertTrue(len(lines)==3)
        self.assertTrue(lines[0]=='Store the following recovery token, in case this record is locked')
        self.assertTrue(lines[2]=='')
        self.assertTrue(len(a2b_base64(lines[1])) == 16 * len(ostore.client.config['servers']))
        stoks = sphinx.split_by_n(a2b_base64(lines[1]), 16)
        for i in range(len(ostore.client.config['servers'])):
          if len(tokens)<max_recovery_tokens and stoks[i] not in tokens[i]:
            tokens[i].add(stoks[i])
          if (len(tokens[i])==max_recovery_tokens):
            self.assertTrue(stoks[i] in tokens[i])

    def test_ostore_recoverytokens_invpwd(self):
      if not ostore.available or ostore_server is None: return
      sys.stdin = Input()
      self.assertIsNone(sphinx.main(('sphinx.py', 'store', user, 'opaque-store.cfg')))

      vp = sphinx.validate_password
      sphinx.validate_password = False

      sys.stdin = Input(pwd="qwer")
      with self.assertRaises(SystemExit) as cm:
          ret = sphinx.main(('sphinx.py', 'recovery-tokens', user))
          self.assertIsNone(ret)
          self.assertEqual(cm.exception.code, 1)

    def test_ostore_unlock(self):
      if not ostore.available or ostore_server is None: return
      sys.stdin = Input()
      f = StringIO()
      with contextlib.redirect_stdout(f):
        self.assertIsNone(sphinx.main(('sphinx.py', 'store', user, 'opaque-store.cfg')))
      lines = f.getvalue().split('\n')
      self.assertTrue(lines[0] == 'successfully created opaque store record. Store the following recovery token, in case this record is locked')
      token = lines[1]

      vp = sphinx.validate_password
      sphinx.validate_password = False

      for i in range(ostore_max_fails+1):
        sys.stdin = Input(pwd="qwer")
        with self.assertRaises(SystemExit) as cm:
            ret = sphinx.main(('sphinx.py', 'read', user))
            self.assertIsNone(ret)
            self.assertEqual(cm.exception.code, 1)

      sys.stdin = Input()
      with self.assertRaises(SystemExit) as cm:
          ret = sphinx.main(('sphinx.py', 'read', user))
          self.assertIsNone(ret)
          self.assertEqual(cm.exception.code, 1)

      time.sleep(3)
      sys.stdin = Input()
      f = StringIO()
      with contextlib.redirect_stdout(f):
        self.assertIsNone(sphinx.main(('sphinx.py', 'unlock', user, token)))
      with open('opaque-store.cfg','r') as fd:
        cfg = fd.read()
      self.assertTrue(cfg in f.getvalue())

    def test_ostore_changepwd(self):
      if not ostore.available or ostore_server is None: return
      sys.stdin = Input()
      self.assertIsNone(sphinx.main(('sphinx.py', 'store', user, 'opaque-store.cfg')))

      sys.stdin = Input(txt='qwer')
      self.assertIsNone(sphinx.main(('sphinx.py', 'changepwd', user)))

      sys.stdin = Input(pwd='qwer')
      f = StringIO()
      with contextlib.redirect_stdout(f):
        self.assertIsNone(sphinx.main(('sphinx.py', 'read', user)))
      with open('opaque-store.cfg','r') as fd:
        cfg = fd.read()
      self.assertTrue(cfg in f.getvalue())

      sys.stdin = Input()
      with self.assertRaises(SystemExit) as cm:
          ret = sphinx.main(('sphinx.py', 'read', user))
          self.assertIsNone(ret)
          self.assertEqual(cm.exception.code, 1)

    def test_ostore_edit(self):
      if not ostore.available or ostore_server is None: return
      sys.stdin = Input()
      self.assertIsNone(sphinx.main(('sphinx.py', 'store', user, 'opaque-store.cfg')))

      sys.stdin = Input()
      environ['EDITOR']=path.dirname(path.abspath(__file__)) + '/editor.py'
      self.assertIsNone(sphinx.main(('sphinx.py', 'edit', user)))

      sys.stdin = Input()
      f = StringIO()
      with contextlib.redirect_stdout(f):
        self.assertIsNone(sphinx.main(('sphinx.py', 'read', user)))
      with open('opaque-store.cfg','r') as fd:
        sorted_cfg = '\n'.join(sorted(fd.read().split('\n')))
      self.assertTrue(sorted_cfg == f.getvalue()[:-1] )

class TestEndToEndNoUserlist(TestEndToEnd):
  def setUp(self):
    if sphinx.userlist:
      sphinx.userlist=False
  def tearDown(self, *args, **kwargs):
    super(TestEndToEndNoUserlist, self).tearDown(*args, **kwargs)
    sphinx.userlist=True

class TestEndToEndNoRWD_Keys(TestEndToEnd):
  def setUp(self):
    if sphinx.rwd_keys:
      sphinx.rwd_keys=False
  def tearDown(self, *args, **kwargs):
    super(TestEndToEndNoRWD_Keys, self).tearDown(*args, **kwargs)
    sphinx.rwd_keys=True

class TestEndToEndNoValidatePassword(TestEndToEnd):
  def setUp(self):
    if sphinx.validate_password:
      sphinx.validate_password=False
  def tearDown(self, *args, **kwargs):
    super(TestEndToEndNoValidatePassword, self).tearDown(*args, **kwargs)
    sphinx.validate_password=True

class TestEndToEndNoneEither(TestEndToEnd):
  def setUp(self):
    if sphinx.validate_password:
      sphinx.validate_password=False
    if sphinx.rwd_keys:
      sphinx.rwd_keys=False
    if sphinx.userlist:
      sphinx.userlist=False
  def tearDown(self, *args, **kwargs):
    super(TestEndToEndNoneEither, self).tearDown(*args, **kwargs)
    sphinx.validate_password=True
    sphinx.rwd_keys=True
    sphinx.userlist=True

class TestEndToEndSingleMode(TestEndToEnd):
  def setUp(self):
    self.threshold=sphinx.threshold
    sphinx.threshold=1
    sphinx.servers=dict(list(sphinx.servers.items())[:1])
    global N
    self.N=N
    N=1

  def tearDown(self, *args, **kwargs):
    super(TestEndToEndSingleMode, self).tearDown(*args, **kwargs)
    sphinx.servers=orig_servers
    sphinx.threshold=self.threshold
    global N
    N=self.N

if __name__ == '__main__':
  unittest.main()




#!/usr/bin/env python3
#
# This file is part of WebSphinx.
#
# SPDX-FileCopyrightText:  2018, Marsiske Stefan <pitchfork@ctrlc.hu>
# SPDX-License-Identifier:  GPL-3.0-or-later
#
# WebSphinx is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation either version 3 of the License, or
# (at your option) any later version.
#
# WebSphinx is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License version 3 for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, If not, see <http://www.gnu.org/licenses/>.


import subprocess
import sys, struct, json, pysodium
import cbor2
from zxcvbn import zxcvbn
from SecureString import clearmem
from pyoprf.multiplexer import Multiplexer
from binascii import b2a_base64
from base64 import urlsafe_b64decode
try:
    from pwdsphinx import sphinx, bin2pass
    from pwdsphinx.config import getcfg
except ImportError:
    import sphinx
    from config import getcfg

cfg = getcfg('sphinx')
pinentry = cfg['websphinx']['pinentry']
log = cfg['websphinx'].get('log')

if 'webauthn_data_dir' not in cfg['websphinx']:
    raise Exception("Cannot find webauthn user directory (webauthn_data_dir). Add it to your config file")

webauthn_data_dir = cfg['websphinx']['webauthn_data_dir']

def handler(cb, cmd, *args):
    m = Multiplexer(sphinx.servers)
    m.connect()
    cb(cmd(m, *args))
    m.close()

def getpwd(title):
    proc=subprocess.Popen([pinentry, '-g'], stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    out, err = proc.communicate(input=('SETTITLE sphinx password prompt\nSETDESC %s\nSETPROMPT master password\ngetpin\n' % (title)).encode())
    if proc.returncode == 0:
        for line in out.split(b'\n'):
            if line.startswith(b"D "): return line[2:]

def fetchOK(proc, cmd):
    proc.stdin.write(f"{cmd}\n".encode("utf8"))
    proc.stdin.flush()
    if((line:=proc.stdout.readline())!=b"OK\n"):
        raise ValueError(f"fail \"{cmd}\": {line}")

def pwdq(pwd):
    q = zxcvbn(pwd.decode('utf8'))
    q['guesses']
    q['score']
    q['crack_times_display']
    q['feedback']

    proc=subprocess.Popen([pinentry, '-g'],
                          stdin=subprocess.PIPE,
                          stdout=subprocess.PIPE,
                          stderr=subprocess.PIPE)

    if not (resp:=proc.stdout.readline()).startswith(b'OK Pleased to meet you'):
        raise ValueError(f"strange greeting \"{resp}\"")
    fetchOK(proc ,"SETTITLE Password Quality Check")
    fetchOK(proc ,"SETOK use this")
    fetchOK(proc ,"SETCANCEL try another")
    fetchOK(proc ,"SETDESC your %s%s (%s/4) master password:%%0a - can be online recovered in %s,%%0a - offline in %s,%%0a - trying ~%s guesses%%0a%%0aAre you sure you want to use this password?" %
            ("★" * q['score'],
             "☆" * (4-q['score']),
             q['score'],
             q['crack_times_display']['online_throttling_100_per_hour'],
             q['crack_times_display']['offline_slow_hashing_1e4_per_second'],
             q['guesses']))
    try:
        fetchOK(proc ,"CONFIRM")
    except ValueError:
        return False
    return True

# Send message using Native messaging protocol
def send_message(data):
  msg = json.dumps(data).encode('utf-8')
  if log:
    log.write(msg)
    log.write(b'\n')
    log.flush()
  length = struct.pack('@I', len(msg))
  sys.stdout.buffer.write(length)
  sys.stdout.buffer.write(msg)
  sys.stdout.buffer.flush()

def users(data):
  def callback(users):
    res = {'names': [i for i in users.split("\n")],
           'cmd': 'list', "mode": data['mode'], 'site': data['site']}
    send_message({ 'results': res })

  try:
    handler(callback, sphinx.users, data['site'])
  except:
    send_message({ 'results': 'fail' })

def get(data):
  def callback(arg):
    res = { 'password': arg, 'name': data['name'], 'site': data['site'], 'cmd': 'login', "mode": data['mode']}
    send_message({ 'results': res })
  try:
    pwd=getpwd("get password for user \"%s\" at host \"%s\"" % (data['name'], data['site']))
    handler(callback, sphinx.get, pwd, data['name'], data['site'])
  except:
    send_message({ 'results': 'fail' })

def create(data):
  def callback(arg):
    res = { 'password': arg, 'name': data['name'], 'site': data['site'], 'cmd': 'create', "mode": data['mode']}
    send_message({ 'results': res })
  try:
    pwd=None
    while not pwd:
        pwd=getpwd("create password for user \"%s\" at host \"%s\"%%0a" % (data['name'], data['site']))
        pwd2=getpwd("REPEAT: create for user \"%s\" at host \"%s\"%%0a" % (data['name'], data['site']))
        if pwd != pwd2:
            send_message({ 'results': 'fail' })
            return
        if not pwdq(pwd): pwd=None

    symbols = ''
    if 's' in data['rules']:
      symbols = bin2pass.symbols
      data['rules'] = ''.join(set(data['rules']) - set(['s']))
    handler(callback, sphinx.create, pwd, data['name'], data['site'], data['rules'], symbols, int(data['size']), None)
  except:
    send_message({ 'results': 'fail' })

def change(data):
  def callback(arg):
    res = { 'password': arg, 'name': data['name'], 'site': data['site'], 'cmd': 'change', "mode": data['mode']}
    send_message({ 'results': res })
  try:
    oldpwd=""
    if cfg['client'].get('rwd_keys'):
        oldpwd=getpwd("current password for \"%s\" at host: \"%s\"%%0a" % (data['name'], data['site']))
    pwd=None
    while not pwd:
        pwd=getpwd("new password for user \"%s\" at host \"%s\"%%0a" % (data['name'], data['site']))
        pwd2=getpwd("REPEAT: new for user \"%s\" at host \"%s\"%%0a" % (data['name'], data['site']))
        if pwd != pwd2:
            send_message({ 'results': 'fail' })
            return
        if not pwdq(pwd): pwd=None

    symbols = ''
    if 's' in data['rules']:
      symbols = bin2pass.symbols
      data['rules'] = ''.join(set(data['rules']) - set(['s']))
    handler(callback, sphinx.change, oldpwd, pwd, data['name'], data['site'], data['rules'], symbols, int(data['size']), None)
  except:
    send_message({ 'results': 'fail' })

def commit(data):
  def callback(arg):
    res = { 'result': arg, 'name': data['name'], 'site': data['site'], 'cmd': 'commit', "mode": data['mode']}
    send_message({ 'results': res })
  try:
    pwd=""
    if cfg['client'].get('rwd_keys'):
      pwd=getpwd("commit password for \"%s\" at host: \"%s\"%%0a" % (data['name'], data['site']))
    handler(callback, sphinx.commit, pwd, data['name'], data['site'])
  except:
    send_message({ 'results': 'fail' })

def undo(data):
  def callback(arg):
    res = { 'result': arg, 'name': data['name'], 'site': data['site'], 'cmd': 'undo', "mode": data['mode']}
    send_message({ 'results': res })
  try:
    pwd=""
    if cfg['client'].get('rwd_keys'):
      pwd=getpwd("undo password for \"%s\" at host: \"%s\"%%0a" % (data['name'], data['site']))
    handler(callback, sphinx.undo, pwd, data['name'], data['site'])
  except:
    send_message({ 'results': 'fail' })

def qrcode(data):
  try:
    sphinx.qrcode("svg", True)
    res = { 'result': arg, 'cmd': 'qrcode', "mode": data['mode']}
    send_message({ 'results': res })
  except:
    send_message({ 'results': 'fail' })


def echo(data):
    send_message(data)

def webauthn_create(data):
  try:
    pwd=None
    while not pwd:
        pwd=getpwd("create password for user \"%s\" at host \"%s\"%%0a" % (data['name'], data['site']))
        pwd2=getpwd("REPEAT: create for user \"%s\" at host \"%s\"%%0a" % (data['name'], data['site']))
        if pwd != pwd2:
            send_message({ 'results': 'fail', 'id': data['id'], 'tabId': data['tabId'], 'cmd': res['cmd']})
            return
        if not pwdq(pwd): pwd=None
    # TODO use webauthn:// instead of raw:// - don't forget to rewrite it in handle()
    rand_bytes = pysodium.randombytes(32)
    pk, sk = pysodium.crypto_sign_seed_keypair(rand_bytes)
    # signed_challenge = pysodium.crypto_sign(data['challenge'], sk)
    auth_data = pysodium.crypto_hash_sha256(data['site'].encode('utf8')) + ( # rpIdHash
        b'\x4d' + # flags https://developer.mozilla.org/en-US/docs/Web/API/Web_Authentication_API/Authenticator_data
        b'\x00' * 4 + # signCount
        b'\x00' * 16 + # attestedCredentialData->AAGUID
        struct.pack('>H', len(pk)) + # attestedCredentialData->pk length
        pk + # attestedCredentialData->credentialId (pk)
        cbor2.dumps({ # cose encoded public key https://datatracker.ietf.org/doc/html/rfc8152#section-13
            1: 1, # key type => okp
            3: -8, # algorithm => Ed25519 (-8)
            -1: 6, # curve => ed25519
            -2: pk, # x => pubkey
            # no y required (y is -3)
    })) # attestedCredentialData->credentialPublicKey

    # TODO use opaque
    with open(webauthn_data_dir + "/" + pk.hex(), "wb") as wf:
        wf.write(urlsafe_b64decode(data['userid']+"=="))

    att_sig =  pysodium.crypto_sign_detached(auth_data + pysodium.crypto_hash_sha256(data['clientDataJSON'].encode('utf-8')), sk)

    attestation = cbor2.dumps({
        'fmt': 'packed',
        'attStmt': {
            'alg': -8,
            'sig': att_sig,
        },
        'authData': auth_data
    })
    #log.write((data['clientDataJSON']+"\n").encode('utf-8'))
    #log.write((auth_data.hex() + "\n").encode('utf-8'))

    res = {
        'pk': b2a_base64(pk).decode('utf8').strip(),
        'attestationObject': b2a_base64(attestation).decode('utf-8').strip(),
        'authData': b2a_base64(auth_data).decode('utf-8').strip(),
        'name': data['name'],
        'site': data['site'],
        'cmd': 'webauthn-create',
        'id': data['id'],
        'tabId': data['tabId']
    }

    # create new user with pk
    m = Multiplexer(sphinx.servers)
    m.connect()
    # TODO add optional argument to create() to skip extending the userlist
    orig_userlist = sphinx.userlist
    sphinx.userlist = False
    sphinx.create(m, pwd, 'raw://'+pk.hex(), data['site'], '', '', target = rand_bytes)
    sphinx.userlist = orig_userlist
    m.close()
    clearmem(sk)
    clearmem(rand_bytes)
    send_message({'results': res})
  except Exception as e:
      send_message({ 'results': 'fail', 'id': data.get('id', ''), 'tabId': data.get('tabId', -1), 'cmd': 'webauthn-create'})
      #send_message({ 'results': 'fail', 'id': data.get('id', ''), 'tabId': data.get('tabId', -1), 'cmd': 'webauthn-create', 'exception': str(e)})

def webauthn_get(data):
    def callback(rand_bytes):
        pk, sk = pysodium.crypto_sign_seed_keypair(rand_bytes)
        clearmem(rand_bytes)

        # TODO use opaque
        with open(webauthn_data_dir + "/" + pk.hex(), "rb") as wf:
            userid = wf.read()

        auth_data = pysodium.crypto_hash_sha256(data['site'].encode('utf8')) + ( # rpIdHash
            b'\x4d' + # flags https://developer.mozilla.org/en-US/docs/Web/API/Web_Authentication_API/Authenticator_data
            b'\x00' * 4 + # signCount
            b'\x00' * 16 + # attestedCredentialData->AAGUID
            struct.pack('>H', len(pk)) + # attestedCredentialData->pk length
            pk + # attestedCredentialData->credentialId (pk)
            cbor2.dumps({ # cose encoded public key https://datatracker.ietf.org/doc/html/rfc8152#section-13
                1: 1, # key type => okp
                3: -8, # algorithm => Ed25519 (-8)
                -1: 6, # curve => ed25519
                -2: pk, # x => pubkey
                # no y required (y is -3)
        })) # attestedCredentialData->credentialPublicKey
        sig = pysodium.crypto_sign_detached(auth_data + pysodium.crypto_hash_sha256(data['clientDataJSON'].encode('utf-8')), sk)
        clearmem(sk)

        res = {
            'userid': b2a_base64(userid).decode('utf-8').strip(),
            'sig': b2a_base64(sig).decode('utf8').strip(),
            'authData': b2a_base64(auth_data).decode('utf-8').strip(),
            'site': data['site'],
            'cmd': 'webauthn-get',
            'id': data['id'],
            'tabId': data['tabId'],
            'pk': b2a_base64(pk).decode('utf-8').strip(),
        }
        send_message({'results': res})
    try:
        pwd=getpwd("get webauthn password at host \"%s\"" % data['site'])
        pk = urlsafe_b64decode(data['pk']+"==")
        handler(callback, sphinx.get, pwd, 'raw://'+pk.hex(), data['site'])
    except Exception as e:
        send_message({ 'results': 'fail', 'id': data.get('id', ''), 'tabId': data.get('tabId', -1), 'cmd': 'webauthn-get'})
        #send_message({ 'results': 'fail', 'id': data.get('id', ''), 'tabId': data.get('tabId', -1), 'cmd': 'webauthn-get', 'exception': str(e)})


func_map = {
    'login': get,
    'list': users,
    'create': create,
    'change': change,
    'commit': commit,
    'undo': undo,
    'qrcode': qrcode,
    'echo': echo,
    'webauthn-create': webauthn_create,
    'webauthn-get': webauthn_get,
}


def main():
  global log
  if log: log = open(log,'ab')
  while True:
    # Read message using Native messaging protocol
    length_bytes = sys.stdin.buffer.read(4)
    if len(length_bytes) == 0:
      return

    length = struct.unpack('i', length_bytes)[0]
    data = json.loads(sys.stdin.buffer.read(length).decode('utf-8'))

    if log:
      log.write(repr(data).encode())
      log.write(b'\n')
      log.flush()

    if data['cmd'] in func_map:
        func_map[data['cmd']](data)
    else:
        send_message({ 'results': 'fail' })

if __name__ == '__main__':
  main()



#!/usr/bin/env python3
# SPDX-FileCopyrightText: 2023, Marsiske Stefan
# SPDX-License-Identifier: GPL-3.0-or-later

import os, tomllib
from pwdsphinx.utils import split_by_n

def getcfg(name):
  paths=[
      # read global cfg
      f'/etc/{name}/config',
      # update with per-user configs
      os.path.expanduser(f"~/.{name}rc"),
      # over-ride with local directory config
      os.path.expanduser(f"~/.config/{name}/config"),
      os.path.expanduser(f"{name}.cfg")
  ]
  config = dict()
  for path in paths:
    try:
        with open(path, "rb") as f:
            data = tomllib.load(f)
    except FileNotFoundError:
        continue
    except tomllib.TOMLDecodeError as ex:
        print(f"error in {path} at {ex}")
        continue
    config.update(data)
  return config


if __name__ == '__main__':
  import sys
  getcfg('sphinx').write(sys.stdout)



#!/usr/bin/env python3

import sys, base64, pysodium, binascii
from pwdsphinx.consts import *

# usage
# getpwd | env/bin/sphinx create minisig://test asdf | pipe2tmpfile minisign -R -s @@keyfile@@ -p /tmp/minisig.pub
# getpwd | env/bin/sphinx get minisig://test asdf | pipe2tmpfile minisign -S -s @@keyfile@@ -m filetosign

"""format is
untrusted comment: sphinx generated minisign key\n
base64(
  4564 0000 4232 <\x00 * 48>
  <keynum 8B unique key identifier>
  <secret key 32B>
  <public key 32B>
  <\x00 * 32>
)

use crypto_sign_ed25519_sk_to_pk(unsigned char *pk, const unsigned char *sk);
to derive pubkey from secret key
"""

def privkey(sk, kid):
    raw = (binascii.unhexlify("456400004232") +
           b'\x00' * 48 +
           kid +
           sk +
           b'\x00' * 32)
    return f"untrusted comment: minisign secret key\n{base64.b64encode(raw).decode('utf8')}"

def pubkey(pk, kid):
    raw = (b'Ed' + kid + pk)
    return f"untrusted comment: minisign public key\n{base64.b64encode(raw).decode('utf8')}"

def convert(rwd, user, host, op, *opts):
    seed=rwd[:32]
    kid=rwd[32:40]
    pk,sk=pysodium.crypto_sign_seed_keypair(seed)
    if op in {CREATE, CHANGE}:
        return pubkey(pk, kid)
    return privkey(sk, kid)

schema = {'minisig': convert}



#!/usr/bin/env python3

import sys, base64, pysodium, binascii, struct
from pwdsphinx.consts import *

# usage
# create key and save pubkey
# getpwd | env/bin/sphinx create ssh-ed25519://test asdf | pipe2tmpfile ssh-keygen -e -f @@keyfile@@ >pubkey
# sign file
# getpwd | env/bin/sphinx get ssh-ed25519://test asdf | pipe2tmpfile ssh-keygen -Y sign -n file -f @@keyfile@@ content.txt > content.txt.sig
# verify file with pubkey
# ssh-keygen -Y check-novalidate -n file -f /tmp/ssh-ed.pubkey -s /tmp/content.txt.sig </tmp/content.txt

from itertools import zip_longest # for Python 3.x
def split_by_n(iterable, n):
    return zip_longest(*[iter(iterable)]*n, fillvalue='')

def privkey(rwd, user, host):
    seed=rwd[:32]
    pk,sk=pysodium.crypto_sign_seed_keypair(seed)

    comment = f"{user}@{host}".encode('utf8')
    comment = struct.pack(">I", len(comment)) + comment

    secrethalf = (
           binascii.unhexlify("a2224bbaa2224bba"               # iv/salt? (Not sure about these 8 bytes)
                              # Here's a repeat of the public key (part of the private key pair)
                              "0000000b"                       # int length = 11
                              "7373682d65643235353139"         # string key type = ssh-ed25519
                              "00000020") +                    # int length = 32
                              # public key payload 32 bytes
                              # probably encoding a point on the ed25519 curve
           pk +
           binascii.unhexlify("00000040") +                    # int length = 64
           # 32 bytes private key payload 1
           sk +                                                # really sk[32] + pk[32]
           comment)                                            # int length + comment as string
    # padding 3 bytes incrementing integers, pads to blocksize 8, starts with "remaining payload"
    padding = bytes(i+1 for i in range(-len(secrethalf)%8))
    secrethalf = secrethalf + padding
    assert len(secrethalf) % 8 == 0

    raw = (binascii.unhexlify("6f70656e7373682d6b65792d763100" # ASCII magic "openssh-key-v1" plus null byte
                              "00000004"                       # int length = 4
                              "6e6f6e65"                       #string cipher = none
                              "00000004"                       # int length = 4
                              "6e6f6e65"                       # string kdfname = none
                              "00000000"                       # int length = 0
                              # zero-length kdfoptions placeholder here
                              "00000001"                       # int number of public keys = 1
                              "00000033"                       # int length first public key = 51 (4 + 11 + 4 + 32)
                              "0000000b"                       # int length = 11
                              "7373682d65643235353139"         # string key type = ssh-ed25519
                              "00000020") +                    # int length = 32
           # public key payload 32 bytes
           # probably encoding a point on the ed25519 curve
           pk +
           struct.pack(">I", len(secrethalf)) +                # int length = 144 size of remaining payload
                                                               # 8 + 4 + 11 + 4 + 32 + 4 + 64 + 4 + {10 + 3}
           secrethalf)

    return ("-----BEGIN OPENSSH PRIVATE KEY-----\n" +
            '\n'.join(''.join(l) for l in split_by_n(base64.b64encode(raw).decode('utf8'), 70)) +
            "\n-----END OPENSSH PRIVATE KEY-----")

def pubkey(rwd, user, host):
    seed=rwd[:32]
    pk,sk=pysodium.crypto_sign_seed_keypair(seed)

    raw = (binascii.unhexlify("0000000b"                       # int length = 11
                              "7373682d65643235353139"         # string key type = ssh-ed25519
                              "00000020") +                    # int length = 32
           pk)

    return f"ssh-ed25519 {base64.b64encode(raw).decode('utf8')} {user}@{host}"

def convert(rwd, user, host, op, *opts):
    if op in {CREATE, CHANGE}:
        return pubkey(rwd, user, host)
    return privkey(rwd,user, host)

schema = {'ssh-ed25519': convert}



#!/usr/bin/env python3

# SPDX-FileCopyrightText: 2018-2024, Marsiske Stefan
# SPDX-License-Identifier: GPL-3.0-or-later

import sys, os, socket, ssl, struct, platform, getpass, binascii, zlib
import concurrent.futures
from SecureString import clearmem
import pysodium, pyoprf
from qrcodegen import QrCode
try:
    from zxcvbn import zxcvbn
except ImportError:
    zxcvbn = None
from equihash import solve
from itertools import permutations
from pyoprf.multiplexer import Peer, Multiplexer as origMultiplexer

try:
  from pwdsphinx import bin2pass, v1sphinx
  from pwdsphinx.config import getcfg
  from pwdsphinx.consts import *
  from pwdsphinx.utils import split_by_n
  from pwdsphinx.ext import init_browser_ext
  from pwdsphinx.converter import convert, convertedBy
  from pwdsphinx import ostore
except ImportError:
  import bin2pass, ostore, v1sphinx
  from config import getcfg
  from consts import *
  from utils import split_by_n
  from ext import init_browser_ext
  from converter import convert, convertedBy

# monkey patching multiplexer that did not set timeouts for peer network
# connections - todo remove this when pyoprf with a fix is suitably deployed
class Multiplexer(origMultiplexer):
    def __init__(self, peers):
        self.peers = [Peer(name
						   ,(p['host'],p['port'])
						   ,type=p.get("type", "SSL")
						   ,ssl_cert = p.get('ssl_cert')
						   ,timeout = p.get('timeout', timeout))
                      for name, p in peers.items()]

win=False
if platform.system() == 'Windows':
  win=True

#### config ####

cfg = getcfg('sphinx')
if 'client' not in cfg:
  print("Error: No configuration found. Check out man sphinx(1) to get started.\n"
        "Look at https://sphinx.pm/servers.html to choose some servers.\n"
        "Continuing with default values, which will probably not work", file=sys.stderr)
debug = cfg.get('client',{}).get('debug', False)
verbose = cfg.get('client',{}).get('verbose', False)
hostname = cfg.get('client',{}).get('address','127.0.0.1')
address = socket.gethostbyname(hostname)
port = int(cfg.get('client',{}).get('port',2355))
timeout = int(cfg.get('client',{}).get('timeout', '5'))
datadir = os.path.expanduser(cfg.get('client',{}).get('datadir','~/.config/sphinx'))
try:
  ssl_cert = os.path.expanduser(cfg.get('client',{}).get('ssl_cert')) # only for dev, production system should use proper certs!
except TypeError: # ignore exception in case ssl_cert is not set, thus None is attempted to expand.
  ssl_cert = None
#  make RWD optional in (sign|seal)key, if it is b'' then this protects against
#  offline master pwd bruteforce attacks, drawback that for known (host,username) tuples
#  the seeds/blobs can be controlled by an attacker if the masterkey is known
rwd_keys = cfg.get('client',{}).get('rwd_keys', False)
validate_password = cfg.get('client',{}).get('validate_password',True)
userlist = cfg.get('client',{}).get('userlist', True)
threshold = int(cfg.get('client',{}).get('threshold') or "1")
ts_epsilon = 1200 # todo make configurable
servers = cfg.get('servers',{})
delete_upgraded = False
if v1sphinx.enabled:
    delete_upgraded = cfg.get('client',{}).get('delete_upgraded',False)

if len(servers)>1:
    if threshold < 2:
        print('if you have multiple servers in your config, you must specify a threshold >1 also')
        exit(1)
    if len(servers)<threshold:
        print(f'threshold({threshold}) must be less or equal than the number of servers({len(servers)}) in your config')
        exit(1)
    for name, server in servers.items():
        if not server.get('timeout'): server['timeout'] = timeout
        if 'ltsigkey' not in server:
           if not 'ltsigkey_path' in server:
               print(f'sphinx server {name} has neither "ltsigkey" nor "ltsigkey_path" configured.\n'
                     f'please obtain the servers public key, and store it as a file at "ltsigkey_path" or base64 encoded at "ltsigkey"')
               exit(1)
           if not os.path.exists(server.get('ltsigkey_path')):
               # deferring errors until usage. maybe never needed, unittests generate these on the fly.
               server['ltsigkey']=None
               continue
           with open(server.get('ltsigkey_path'),'rb') as fd:
              ltsigkey = fd.read()
              if(len(ltsigkey)!=pysodium.crypto_sign_PUBLICKEYBYTES):
                 raise ValueError(f"long-term signature key for server {name} at {server['ltsigkey_path']} is of incorrect size")
           server['ltsigkey']=ltsigkey
        else:
            server['ltsigkey'] = binascii.a2b_base64(server['ltsigkey'])
            if 'ltsigkey_path' in server:
               # oh we have both?
               if not os.path.exists(server.get('ltsigkey_path')):
                  print(f"server {name} has both ltsigkey and ltsigkey_path set, but the latter file doesn't exist, check your assumptions.")
                  continue
               with open(server.get('ltsigkey_path'),'rb') as fd:
                  ltsigkey = fd.read()
                  if(len(ltsigkey)!=pysodium.crypto_sign_PUBLICKEYBYTES):
                     raise ValueError(f"long-term signature key for server {name} at {server['ltsigkey_path']} is of incorrect size")
               if ltsigkey!=server['ltsigkey']:
                  print(f'sphinx server {name} has both "ltsigkey" and "ltsigkey_path" configured, and they are different!\n'
                        f'please obtain the servers public key, and store it as a file at "ltsigkey_path" or base64 encoded aat "ltsigkey"\n'
                        f"and don't forget to delete the other variable")
                  exit(1)
elif threshold > 1:
    print(f'threshold({threshold}) must be less or equal than the number of servers({len(servers)}) in your config')
    exit(1)

if verbose:
    print("hostname:", hostname, file=sys.stderr)
    print("address:", address, file=sys.stderr)
    print("port:", port, file=sys.stderr)
    print("datadir:", datadir, file=sys.stderr)
    print("ssl_cert:", ssl_cert, file=sys.stderr)
    print("rwd_keys:", rwd_keys, file=sys.stderr)
    print("validate_password:", validate_password, file=sys.stderr)
    print("userlist:", userlist, file=sys.stderr)
    print("threshold:", threshold, file=sys.stderr)
    for name, server in servers.items():
      print(f"{name} {server.get('host','localhost')}:{server.get('port', 2355)} {server['ltsigkey']} cert: {server.get('ssl_cert')}")

#### consts ####

ENC_CTX = b"sphinx encryption key"
SIGN_CTX = b"sphinx signing key"
SALT_CTX = b"sphinx host salt"
PASS_CTX = b"sphinx password context"
CHECK_CTX = b"sphinx check digit context"

#### Helper fns ####

def get_masterkey():
  try:
    with open(os.path.join(datadir,'masterkey'), 'rb') as fd:
        mk = fd.read()
    return mk
  except FileNotFoundError:
    raise ValueError("ERROR: Could not find masterkey!\nIf sphinx was working previously it is now broken.\nIf this is a fresh install all is good, you just need to run `%s init`." % sys.argv[0])

def get_signkey(id, rwd):
  mk = get_masterkey()
  seed = pysodium.crypto_generichash(SIGN_CTX, mk)
  clearmem(mk)
  # rehash with rwd so the user always contributes his pwd and the sphinx server it's seed
  seed = pysodium.crypto_generichash(seed, id)
  if rwd_keys:
    seed = pysodium.crypto_generichash(seed, rwd)
  pk, sk = pysodium.crypto_sign_seed_keypair(seed)
  clearmem(seed)
  return sk, pk

def get_sealkey():
  mk = get_masterkey()
  sk = pysodium.crypto_generichash(ENC_CTX, mk)
  clearmem(mk)
  return sk

def encrypt_blob(blob):
  # todo implement padding to hide length information
  sk = get_sealkey()
  nonce = pysodium.randombytes(pysodium.crypto_secretbox_NONCEBYTES)
  ct = pysodium.crypto_aead_xchacha20poly1305_ietf_encrypt(blob,VERSION,nonce,sk)
  clearmem(sk)
  return VERSION+nonce+ct

def decrypt_blob(blob):
  # todo implement padding to hide length information
  sk = get_sealkey()
  version = blob[:1]
  if version > VERSION:
    raise ValueError("Your client is too old to handle this response. Please update your client.")
  blob = blob[1:]
  nonce = blob[:pysodium.crypto_secretbox_NONCEBYTES]
  blob = blob[pysodium.crypto_secretbox_NONCEBYTES:]
  res = pysodium.crypto_aead_xchacha20poly1305_ietf_decrypt(blob,version,nonce,sk)
  clearmem(sk)
  return version, res

def sign_blob(blob, id, rwd):
  sk, pk = get_signkey(id, rwd)
  res = pysodium.crypto_sign_detached(blob,sk)
  clearmem(sk)
  return b''.join((blob,res))

def getsalt():
  mk = get_masterkey()
  salt = pysodium.crypto_generichash(SALT_CTX, mk)
  clearmem(mk)
  return salt

def getid1(host,user,peer,salt=None):
  if salt is None: salt = getsalt()
  user_len = struct.pack('!H', len(user))
  host_len = struct.pack('!H', len(host))
  peer_len = struct.pack('!H', len(peer))
  return pysodium.crypto_generichash(b''.join((user_len,user.encode(),
                                               host_len,host.encode(),
                                               peer_len,peer.encode())), salt, 32)
def getid(host, user, m):
  salt = getsalt()
  ids = []
  for peer in m:
    ids.append(getid1(host,user,peer.name,salt))
  return ids

def oprf2rwd(r, beta, pwd, host, user):
  rwd = pyoprf.unblind_finalize(r, beta, pwd)
  salt = pysodium.crypto_generichash(getsalt()
                                     + struct.pack('!H', len(user.encode('utf8')))
                                     + user.encode('utf8')
                                     + struct.pack('!H', len(host.encode('utf8')))
                                     + host.encode('utf8'),
                                     outlen=pysodium.crypto_pwhash_SALTBYTES)
  return pysodium.crypto_pwhash(pyoprf.OPRF_BYTES, rwd, salt, pysodium.crypto_pwhash_OPSLIMIT_INTERACTIVE, pysodium.crypto_pwhash_MEMLIMIT_INTERACTIVE)

def unpack_rule(ct):
  version, packed = decrypt_blob(ct)
  xor_mask = packed[-64:]
  v = int.from_bytes(packed[:-64], "big")

  size = v & ((1<<7) - 1)
  rule = {c for i,c in enumerate(('u','l','d')) if (v >> 7) & (1 << i)}
  symbols = [c for i,c in enumerate(bin2pass.symbols) if (v>>(7+3) & (1<<i))]
  if validate_password:
      check_digit = (v>>(7+3+33))
  else:
      check_digit = 0

  return rule, symbols, size, check_digit, xor_mask

def pack_rule(char_classes, syms, size, check_digit, xor_mask=None, with_schema=False):
  # pack rules into and encrypt them
  if set(char_classes) - {'u','l','d'}:
    raise ValueError("ERROR: rules can only contain any of 'uld'.")
  if set(syms) - set(bin2pass.symbols) != set():
    raise ValueError("ERROR: symbols can only contain any of '%s'." % bin2pass.symbols)
  if not with_schema and (char_classes == '' and len(syms)<2):
    raise ValueError("ERROR: no char classes and not enough symbols specified.")
  if not with_schema and (xor_mask is None and (char_classes == '' and len(syms)<2)):
    raise ValueError("ERROR: no char classes and not enough symbols specified.")
  if xor_mask is None:
      xor_mask = b'\x00' * 64
  elif len(xor_mask)!=64:
    raise ValueError("ERROR: xor_mask must be 64bytes, is instead: %d." % len(xor_mask))
  if size<0 or size>127:
    raise ValueError("ERROR: invalid max password size: %d." % size)

  packed = size
  packed = packed + (sum(1<<i for i, c in enumerate(('u','l','d')) if c in char_classes) << 7)
  packed = packed + (sum(1<<i for i, c in enumerate(bin2pass.symbols) if c in syms) << (7 + 3))
  packed = packed + ((check_digit & ((1<<5) - 1)) << (7 + 3 + 33) )
  pt = packed.to_bytes(6,"big") + xor_mask
  return encrypt_blob(pt)

def xor(x,y):
  return bytes(a ^ b for (a, b) in zip(x, y))

def commit_undo(m, op, pwd, user, host):
  ids = getid(host, user,m)
  r, alpha = pyoprf.blind(pwd)
  msgs = [b''.join([op, id, alpha]) for id in ids]
  m = ratelimit(m, msgs, len(ids))
  if not auth(m,msgs,ids,host,user,alpha,pwd,r):
    m.close()
    raise ValueError("Failed to authenticate to server while %s" % "committing" if op == COMMIT else "undoing")
  if set(m.gather(2))!={b'ok'}:
    m.close()
    raise ValueError("Server failed to %s" % "Commit" if op == COMMIT else "UNDO")
  m.close()
  return True

def read_pkt(s,size):
    res = []
    read = 0
    while read<size or len(res[-1])==0:
      res.append(s.recv(size-read))
      read+=len(res[-1])

    return b''.join(res)

def update_rec(s, id, item): # this is only for user blobs. a UI feature offering a list of potential usernames.
    if not userlist:
        s.send(b"\0"*96)
        return
    signed_id = sign_blob(id, id, b'')
    s.send(signed_id)
    # wait for user blob
    bsize = s.recv(2)
    bsize = struct.unpack('!H', bsize)[0]
    if bsize == 0:
      # it is a new blob, we need to attach an auth signing pubkey
      sk, pk = get_signkey(id, b'')
      clearmem(sk)
      # we encrypt with an empty rwd, so that the user list is independent of the master pwd
      blob = encrypt_blob(item.encode())
      bsize = len(blob)
      if bsize >= 2**16:
        s.close()
        raise ValueError("ERROR: list of usernames is bigger than 64KB. %d" % bsize)
      blob = struct.pack("!H", bsize) + blob
      # writes need to be signed, and sinces its a new blob, we need to attach the pubkey
      blob = b''.join([pk, blob])
      # again no rwd, to be independent of the master pwd
      blob = sign_blob(blob, id, b'')
    else:
      blob = read_pkt(s, bsize)
      if blob == b'fail':
        s.close()
        raise ValueError("reading list of user names failed")
      version, blob = decrypt_blob(blob)
      items = {x for x in blob.decode().split('\x00') if x}
      # this should not happen, but maybe it's a sign of corruption?
      if item in items:
        print(f'warning: "{item}" is already in the user record', file=sys.stderr)
      items.add(item)
      blob = ('\x00'.join(sorted(items))).encode()
      # notice we do not add rwd to encryption of user blobs
      blob = encrypt_blob(blob)
      bsize = len(blob)
      if bsize >= 2**16:
        s.close()
        raise ValueError("ERROR: list of user names is bigger than 64KB. %d" % bsize)
      blob = struct.pack("!H", bsize) + blob
      blob = sign_blob(blob, id, b'')
    s.send(blob)

def auth(m,ops,ids,host,user,alpha=None,pwd=None,r=None):
  if r is None:
    nonces = m.gather(32)
    if nonces is None:
      m.close()
      return False
    nonces = list(enumerate(nonces))
    rwd = b''
  else:
    msgs = m.gather(65,proc=lambda x: (x[:33],x[33:]))
    if msgs is None:
      m.close()
      return False
    nonces = [(idx,resp[1]) for idx, resp in enumerate(msgs) if resp is not None]
    if sum(1 for m in msgs if m is not None) < len(m):
        raise ValueError("auth: not all peers answered or authenticated")
    if len(servers)>1:
        beta = pyoprf.thresholdmult([resp[0] for resp in msgs if resp is not None][:threshold])
    else:
        beta = msgs[0][0][1:]
    #rwd = pyoprf.unblind_finalize(r, beta, pwd)
    rwd = oprf2rwd(r, beta, pwd, host, user)

  for (idx, nonce), op in zip(nonces, ops):
    sk, pk = get_signkey(ids[idx], rwd)
    sig = pysodium.crypto_sign_detached(op+nonce,sk)
    clearmem(sk)
    m.send(idx, sig)

  clearmem(rwd)

  responses = m.gather(6)
  if responses is None:
    m.close()
    return False

  fails = 0
  for idx, resp in enumerate(responses):
    if resp==b'\x00\x04auth': continue
    if resp==b'\x00\x04fail':
      print(f'authentication failed for {m[idx].name}')
      m[idx].close()
      fails+=1
    else:
      raise ValueError("unexpected auth result")

  if fails > 0:
      raise ValueError("some peers have failed to authenticate us")

  return True

def ratelimit_step1(m, reqs, min_conns):
  for i, req in enumerate(reqs):
    pkt0 = b''.join([CHALLENGE_CREATE, req])
    m[i].send(pkt0)

  challenges = m.gather(1+1+8+32, n=min_conns) # n,k,ts,sig
  if challenges is None:
    m.close()
    return False
  m.close()
  return challenges

def ratelimit(m,reqs,min_conns):
  challenges  = ratelimit_step1(m, reqs, min_conns)

  puzzles = []
  with concurrent.futures.ProcessPoolExecutor() as executor:
    for idx, challenge in enumerate(challenges):
        if challenge in {None, b''}: continue
        n = challenge[0]
        k = challenge[1]

        try:
            os.write(3,f"{idx} {n} {k}\n".encode('utf8'))
        except OSError: pass

        if k==4:
            if n < 90:
              if verbose: print(f"{m[idx].name} sent an easy puzzle: %d" % n, file=sys.stderr)
            elif n > 100:
              if verbose: print(f"{m[idx].name} sent a hard puzzle: %d" % n, file=sys.stderr)
            else:
              if verbose: print(f"{m[idx].name} sent a moderate puzzle: %d" % n, file=sys.stderr)
        seed = challenge + reqs[idx]

        solution = executor.submit(solve,n,k,seed)
        puzzles.append((challenge, solution, idx))

  for puzzle in puzzles:
    m[puzzle[2]].connect()
    pkt1 = b''.join([CHALLENGE_VERIFY, puzzle[0]])
    m.send(puzzle[2], pkt1)
    m.send(puzzle[2], reqs[puzzle[2]])
    m.send(puzzle[2], puzzle[1].result())
  return m

def getpwd():
  if sys.stdin.isatty():
    return getpass.getpass("enter your password please: ").encode('utf8')
  else:
    return sys.stdin.buffer.readline().rstrip(b'\n')

def dkg(m, op, threshold, keyids, alpha):
   n = len(m)

   # load peer long-term keys
   peer_lt_pks = []
   for name, server in servers.items():
      k = server.get('ltsigkey')
      if k is None:
        print(f"Warning the ltsigkey for server: {name} is unset.")
        continue
      peer_lt_pks.append(k)
   if len(peer_lt_pks) < len(servers.items()):
       raise ValueError("cannot create record in threshold setup without having the long-term signing keys of all servers")

   if op == CREATE_DKG:
     tp, msg0 = pyoprf.tpdkg_start_tp(n, threshold, ts_epsilon, "threshold sphinx dkg create k", peer_lt_pks)
     for index, id in enumerate(keyids):
        msg = b"%c%s%s%s" % (CREATE_DKG, id, alpha, msg0)
        m.send(index,msg)
   else:
     tp, msg0 = pyoprf.tpdkg_start_tp(n, threshold, ts_epsilon, "threshold sphinx dkg change k", peer_lt_pks)
     msg = b"%s%s" % (alpha, msg0)
     m.broadcast(msg)

   while pyoprf.tpdkg_tp_not_done(tp):
     ret, sizes = pyoprf.tpdkg_tp_input_sizes(tp)
     peer_msgs = []
     if ret:
         if sizes[0] > 0:
             #print(f"step: {tp[0].step}")
             peer_msgs = m.gather(sizes[0],n) #,debug=True)
     else:
         peer_msgs = [m[i].read(s) if s>0 else b'' for i, s in enumerate(sizes)]
     msgs = b''.join(peer_msgs)

     cur_step = pyoprf.tpdkg_tpstate_step(tp)
     try:
       out = pyoprf.tpdkg_tp_next(tp, msgs)
     except Exception as e:
       m.close()
       if pyoprf.tpdkg_tpstate_cheater_len(tp) > 0:
         cheaters, cheats = pyoprf.tpdkg_get_cheaters(tp)
         msg=[f"Warning during the distributed key generation the peers misbehaved: {sorted(cheaters)}"]
         for k, v in cheats:
           msg.append(f"\tmisbehaving peer: {k} was caught: {v}")
         msg = '\n'.join(msg)
         raise ValueError(msg)
       else:
         raise ValueError(f"{e} | tp step {cur_step}")
     if(len(out)>0):
       for i in range(pyoprf.tpdkg_tpstate_n(tp)):
         msg = pyoprf.tpdkg_tp_peer_msg(tp, out, i)
         m.send(i, msg)

   betas = m.gather(33, n)
   if betas is None:
     m.close()
     raise ValueError(f"failed to get oprf responses from shareholders in final step of dkg")

   rwds = set(pyoprf.thresholdmult([betas[i] for i in order])
              for order in permutations(range(n),threshold))
   if len(rwds) != 1:
     raise ValueError("DKG shares are inconsistent, aborting operation.")

   return list(rwds)[0]

#### OPs ####

def create_masterkey():
  kfile = os.path.join(datadir,'masterkey')
  if os.path.exists(kfile):
    print("Already initialized.", file=sys.stderr)
    return 1
  if not os.path.exists(datadir):
    os.makedirs(datadir, 0o700, exist_ok=True)
  mk = pysodium.randombytes(32)
  try:
    with open(kfile,'wb') as fd:
      if not win: os.fchmod(fd.fileno(),0o600)
      fd.write(mk)
  except:
    print("ERROR: failed to initialize master key", file=sys.stderr)
    return 1
  finally:
    print("you want to make a backup of the following masterkey, keep it safe and secure", binascii.b2a_base64(mk))
    print('this key is also stored - and must be available - at:', kfile)
    clearmem(mk)

def init():
  init_browser_ext()
  create_masterkey()
  # create health check record
  m = Multiplexer(servers)
  m.connect()
  create(m, b"all ok?", "healthcheck", "sphinx servers", target="everything works fine")
  return 0

def create(m, pwd, user, host, char_classes='uld', symbols=bin2pass.symbols, size=0, target=None):
  # 1st step OPRF on the new seed
  ids = getid(host, user, m)
  r, alpha = pyoprf.blind(pwd)
  if threshold > 1:
    beta = dkg(m, CREATE_DKG, threshold, ids, alpha)
  else:
    msg = b''.join([CREATE, ids[0], alpha])
    m.broadcast(msg)

    # wait for response from sphinx server
    beta = m.gather(33)
    if beta is None:
      m.close()
      raise ValueError("ERROR: Creating new password, the record probably already exists or the first message to server was corrupted during transport.")
      # or (less probable) the initial message was longer/shorter than the 65 bytes we sent
      # or (even? less probable) the value alpha received by the server is not a valid point
      # both of these less probable causes point at corruption during transport
    beta = beta[0][1:]

  #rwd = pyoprf.unblind_finalize(r, beta, pwd)
  rwd = oprf2rwd(r, beta, pwd, host, user)

  # second phase, derive new auth signing pubkey
  sign_keys=[]
  for id in ids:
    sk, pk = get_signkey(id, rwd)
    clearmem(sk)
    sign_keys.append(pk)

  if validate_password:
      checkdigit = pysodium.crypto_generichash(CHECK_CTX, rwd, 1)[0]
  else:
      checkdigit = 0

  if target and not user.startswith('raw://'):
    trwd, char_classes, symbols = bin2pass.pass2bin(target, None)
    xormask = xor(pysodium.crypto_generichash(PASS_CTX, rwd, outlen=64),trwd)
    size = len(target)
    #char_classes = 'uld'
    #symbols = bin2pass.symbols
  elif target and user.startswith('raw://'):
    size = len(target)
    target = target + pysodium.randombytes(64 - len(target))
    xormask = xor(pysodium.crypto_generichash(PASS_CTX, rwd, outlen=64),target)
  elif convertedBy(user) is not None:
    xormask = pysodium.randombytes(64)
  else:
    _uppers = set([x.decode('utf8') for x in bin2pass.sets['u']])
    _lowers = set([x.decode('utf8') for x in bin2pass.sets['l']])
    _digits = set([x.decode('utf8') for x in bin2pass.sets['d']])
    _symbols = set(symbols)
    while True:
        xormask = pysodium.randombytes(64)
        candidate = convert(
            xor(pysodium.crypto_generichash(PASS_CTX, rwd, outlen=64),xormask),
            user, host, CREATE,
            char_classes,size,symbols)
        if 1 <= size < 8: break # too much of a bias especially for ulsd when size < 5
        if 'u' in char_classes and len(_uppers.intersection(candidate)) == 0:
            continue
        if 'l' in char_classes and len(_lowers.intersection(candidate)) == 0:
            continue
        if 'd' in char_classes and len(_digits.intersection(candidate)) == 0:
            continue
        if len(_symbols) > 0 and len(_symbols.intersection(candidate)) == 0:
            continue
        break

  rule = pack_rule(char_classes, symbols, size, checkdigit, xormask, with_schema='://' in user)
  # send over new signed(pubkey, rule)
  for i,(id,pk) in enumerate(zip(ids, sign_keys)):
    msg = b''.join([pk, rule])
    msg = sign_blob(msg, id, rwd)
    m[i].send(msg)

  # add user to user list for this host
  # a malicous server could correlate all accounts on this services to this users here
  for p in m:
     id = getid1(host, '', p.name)
     update_rec(p.fd, id, user)
     p.close()

  rwd = xor(pysodium.crypto_generichash(PASS_CTX, rwd, outlen=64),xormask)

  ret = convert(rwd,user,host,CREATE,char_classes,size,symbols)
  clearmem(rwd)
  return ret

def try_v1get(pwd, host, user):
   rwd, classes, size, symbols = v1sphinx.get(pwd, user, host)
   ret = convert(rwd,user,GET,host,classes,size,symbols)
   if not user.startswith("otp://"):
       target = ret
   else:
       target = bin2pass.derive(rwd, classes, size, symbols)
   clearmem(rwd)
   # lift to v2
   m = Multiplexer(servers)
   m.connect()
   cret = create(m, pwd, user, host, target=target)
   assert ret == cret
   print(f"updated v1 record for {user}@{host} to v2", file=sys.stderr)
   if delete_upgraded:
       try:
           v1sphinx.delete(pwd, user, host)
       except:
           print(f"failed to delete v1 record for {user}@{host}", file=sys.stderr)
       else:
           print(f"deleted v1 record for {user}@{host} record after update to v2", file=sys.stderr)
   return ret

def get(m, pwd, user, host):
  ids = getid(host, user, m)
  r, alpha = pyoprf.blind(pwd)

  msgs = [b''.join([GET, id, alpha]) for id in ids]
  try:
      m = ratelimit(m, msgs, threshold)
  except ValueError:
    if v1sphinx.enabled: return try_v1get(pwd, host, user)
    raise

  connected = sum([1 for p in m if p.state == 'connected'])
  if connected < threshold:
    raise ValueError(f"Failed to get enough shareholders to respond: {connected} responded")

  if len(servers) > 1:
    try:
      resps = m.gather(33+RULE_SIZE, threshold)
    except ValueError:
      if v1sphinx.enabled: return try_v1get(pwd, host, user)
      raise
    if resps is None:
      m.close()
      raise ValueError("Failed to get any answers from shareholders")
    if len({x for x in resps if x is not None}) == 1 and {x for x in resps if x is not None} == {b'\x00\x04fail'}:
      raise ValueError("ERROR: The record does not exist, there's a chance you are being fished.")
    resps = [(x[:33], x[33:]) for x in resps if x is not None]
    if len({resp[1] for resp in resps if resp}) != 1:
      m.close()
      raise ValueError("ERROR: servers disagree on rules")
    rules = resps[0][1]
    beta = pyoprf.thresholdmult([resp[0] for resp in resps if resp])
  else:
    try:
        resp = m.gather(33+RULE_SIZE, 1)[0] # beta + sealed rules
    except ValueError:
      if v1sphinx.enabled: return try_v1get(pwd, host, user)
      raise
    if resp is None:
      m.close()
      raise ValueError("Failed to get answers from sphinx server")
    if resp == b'\x00\x04fail':
      raise ValueError("ERROR: The record does not exist, there's a chance you are being fished.")
    if len(resp)!=33+RULE_SIZE:
      m.close()
      raise ValueError("ERROR: the request to server was corrupted during transport.")
    beta = resp[1:33]
    rules = resp[33:]

  #rwd = pyoprf.unblind_finalize(r, beta, pwd)
  rwd = oprf2rwd(r, beta, pwd, host, user)

  m.close()
  try:
    classes, symbols, size, checkdigit, xormask = unpack_rule(rules)
  except ValueError:
    m.close()
    raise ValueError("ERROR: failed to unpack password rules from server")

  if validate_password and (checkdigit != (pysodium.crypto_generichash(CHECK_CTX, rwd, 1)[0] & ((1<<5)-1))):
    m.close()
    raise ValueError("ERROR: bad checkdigit")

  rwd = xor(pysodium.crypto_generichash(PASS_CTX, rwd, outlen=64),xormask)

  ret = convert(rwd,user,GET,host,classes,size,symbols)
  clearmem(rwd)

  return ret

def read_blob(m, ids, host, rwd = b''):
  msgs = [b''.join([READ, id]) for id in ids]
  m = ratelimit(m, msgs, threshold)

  if auth(m,msgs,ids,host,'') is False:
    m.close()
    return

  bsizes = set(m.gather(2, proc = lambda x: struct.unpack('!H', x)[0]))
  if bsizes is None:
    m.close()
    raise ValueError("failed to get sizes for user blobs from threshold servers")
  if len(bsizes) != 1:
    m.close()
    raise ValueError(f"ERROR: inconsistent user list blob sizes: {bsizes}")
  bsize = list(bsizes)[0]
  #print('got all blobsizes')
  if bsize == 0:
    # this should not happen, it means something is corrupt
    m.close()
    raise ValueError("ERROR: server has no associated user record for this host", file=sys.stderr)

  blobs = m.gather(bsize)
  if blobs is None:
    m.close()
    raise ValueError("failed to read user blobs from sphinx servers")
  #print('got all blobs')
  ptblobs = set()
  for blob in blobs:
    if blob == b'fail':
      m.close()
      raise ValueError("ERROR: invalid signature on list of users")
    ptblobs.add(decrypt_blob(blob))

  if len(ptblobs)!=1:
    raise ValueError(f"ERROR: inconsistent user list blobs")
  blob = list(ptblobs)[0]

  return blob

def users(m, host):
  _, users = read_blob(m, getid(host, '', m), host) or (None, set())
  if users:
      users = set(users.decode().split('\x00'))
  if v1sphinx.enabled:
      try: v1users = v1sphinx.users(host)
      except: v1users = set()
      users = users | v1users
  if not users:
      return "no users found"
  return '\n'.join(sorted(users))

def change(m, oldpwd, newpwd, user, host, char_classes='uld', symbols=bin2pass.symbols, size=0, target=None):
  ids = getid(host, user, m)
  r, alpha = pyoprf.blind(oldpwd)

  if threshold > 1:
    msgs = [b''.join([CHANGE_DKG, id, alpha]) for id in ids]
  else:
    msgs = [b''.join([CHANGE, id, alpha]) for id in ids]
  m = ratelimit(m, msgs, len(ids))
  # auth: do sphinx with current seed, use it to sign the nonce
  if not auth(m,msgs,ids,host,user,alpha,oldpwd,r):
    m.close()
    raise ValueError("ERROR: Failed to authenticate using old password to server while changing password on server or record doesn't exist")

  r, alpha = pyoprf.blind(newpwd)
  if threshold > 1:
    beta = dkg(m, CHANGE_DKG, threshold, None, alpha)
  else:
    m.broadcast(alpha)
    beta = m.gather(33,1) # beta
    if beta is None or len(beta[0])!=33:
      m.close()
      raise ValueError("ERROR: changing password failed due to corruption during transport.")
    beta = beta[0][1:]
  #rwd = pyoprf.unblind_finalize(r, beta, newpwd)
  rwd = oprf2rwd(r, beta, newpwd, host, user)

  if validate_password:
      checkdigit = pysodium.crypto_generichash(CHECK_CTX, rwd, 1)[0]
  else:
      checkdigit = 0

  if target and not user.startswith('raw://'):
    trwd, char_classes, symbols = bin2pass.pass2bin(target, None)
    xormask = xor(pysodium.crypto_generichash(PASS_CTX, rwd, outlen=64),trwd)
    size = len(target)
    #char_classes = 'uld'
    #symbols = bin2pass.symbols
  elif target and user.startswith('raw://'):
    size = len(target)
    target = target + pysodium.randombytes(64 - len(target))
    xormask = xor(pysodium.crypto_generichash(PASS_CTX, rwd, outlen=64),target)
  elif convertedBy(user) is not None:
    xormask = pysodium.randombytes(64)
  else:
    _uppers = set([x.decode('utf8') for x in bin2pass.sets['u']])
    _lowers = set([x.decode('utf8') for x in bin2pass.sets['l']])
    _digits = set([x.decode('utf8') for x in bin2pass.sets['d']])
    _symbols = set(symbols)
    while True:
        xormask = pysodium.randombytes(64)
        candidate = convert(
            xor(pysodium.crypto_generichash(PASS_CTX, rwd, outlen=64),xormask),
            user, host, CREATE,
            char_classes,size,symbols)
        if 1 <= size < 8: break # too much of a bias especially for ulsd when size < 5
        if 'u' in char_classes and len(_uppers.intersection(candidate)) == 0:
            continue
        if 'l' in char_classes and len(_lowers.intersection(candidate)) == 0:
            continue
        if 'd' in char_classes and len(_digits.intersection(candidate)) == 0:
            continue
        if len(_symbols) > 0 and len(_symbols.intersection(candidate)) == 0:
            continue
        break

  rule = pack_rule(char_classes, symbols, size, checkdigit, xormask)

  for i, id in enumerate(ids):
    sk, pk = get_signkey(id, rwd)
    clearmem(sk)
    # send over new signed(pubkey)
    m[i].send(sign_blob(b''.join([pk,rule]), id, rwd))

  if set(m.gather(2))!={b'ok'}:
    m.close()
    raise ValueError("ERROR: failed to update password rules on the server during changing of password.")

  m.close()
  rwd = xor(pysodium.crypto_generichash(PASS_CTX, rwd, outlen=64),xormask)
  ret = convert(rwd,user,host,CHANGE,char_classes,size,symbols)
  clearmem(rwd)

  return ret

def commit(s, pwd, user, host):
  return commit_undo(s, COMMIT, pwd, user, host)

def undo(s, pwd, user, host):
  return commit_undo(s, UNDO, pwd, user, host)

def delete(m, pwd, user, host):
  # run sphinx to recover rwd for authentication
  ids = getid(host, user, m)
  r, alpha = pyoprf.blind(pwd)
  msgs = [b''.join([DELETE, id, alpha]) for id in ids]
  m = ratelimit(m, msgs, len(ids))
  #print("solved ratelimit puzzles")
  if auth(m,msgs,ids,host,user,alpha,pwd,r) is False:
    m.close()
    raise ValueError("ERROR: Failed to authenticate to server while deleting password on server or record doesn't exist")
  #print("authenticated")

  if not userlist:
     m.broadcast(b"\0"*96)
  else:
     # delete user from user list for this host
     # a malicous server could correlate all accounts on this services to this users here
     # first query user record for this host
     for p in m:
        id = getid1(host, '', p.name)
        signed_id = sign_blob(id, id, b'')
        p.send(signed_id)

     # wait for user blob
     bsizes = set(m.gather(2, proc = lambda x: struct.unpack('!H', x)[0]))
     if bsizes is None:
       m.close()
       raise ValueError("failed to get sizes for user blobs from threshold servers")
     if len(bsizes) != 1:
       m.close()
       raise ValueError(f"ERROR: inconsistent user list blob sizes: {bsizes}")
     bsize = list(bsizes)[0]
     #print('got all blobsizes')
     if bsize == 0:
       # this should not happen, it means something is corrupt
       m.close()
       raise ValueError("ERROR: server has no associated user record for this host", file=sys.stderr)

     blobs = m.gather(bsize)
     if blobs is None:
       m.close()
       raise ValueError("failed to read user blobs from sphinx servers")
     #print('got all blobs')
     ptblobs = set()
     for blob in blobs:
       if blob == b'fail':
         m.close()
         raise ValueError("ERROR: invalid signature on list of users")
       ptblobs.add(decrypt_blob(blob))

     if len(ptblobs)!=1:
       raise ValueError(f"ERROR: inconsistent user list blobs")
     version, blob = list(ptblobs)[0]

     users = set(blob.decode().split('\x00'))
     if user not in users:
       # this should not happen, but maybe it's a sign of corruption?
       m.close()
       raise ValueError(f'warning "{user}" is not in user record', file=sys.stderr)
     users.remove(user)
     blob = ('\x00'.join(sorted(users))).encode()
     # notice we do not add rwd to encryption of user blobs
     for p in m:
       xblob = encrypt_blob(blob)
       bsize = len(xblob)
       if bsize >= 2**16:
           m.close()
           raise ValueError("ERROR: blob is bigger than 64KB.")
       xblob = struct.pack("!H", bsize) + xblob
       id = getid1(host, '', p.name)
       xblob = sign_blob(xblob, id, b'')
       #print(f'updating {p.name}\t{xblob.hex()}')
       p.send(xblob)

     if set(m.gather(2))!={b'ok'}:
       m.close()
       raise ValueError("ERROR: server failed to save updated list of user names for host: %s." % host)

  m.close()
  return True

def print_qr(qrcode: QrCode) -> None:
  chars = {
    (True, True):    ' ',      # empty
    (False, True):   '\u2580', # upper
    (True, False):   '\u2584', # lower
    (False, False):  '\u2588', # full
  }
  border = 1
  for y in range(-border, qrcode.get_size() + border, 2):
    for x in range(-border, qrcode.get_size() + border):
      print(chars[(qrcode.get_module(x,y),qrcode.get_module(x,y+1))], end="")
    print()
  print()

def healthcheck(m):
  pwd = b"all ok?"
  user = "healthcheck"
  host = "sphinx servers"
  ids = getid(host, user, m)
  r, alpha = pyoprf.blind(pwd)
  msgs = [b''.join([GET, id, alpha]) for id in ids]
  return ratelimit_step1(m, msgs, len(ids)) != False

def qrcode(output, key):
  mk=get_masterkey() if key else b''
  hosts = []
  for name, server in servers.items():
    if len(name)>255: raise ValueError(f"server name: {name} is too long, max 255 allowed in qr codes")
    hosts.append(bytes([len(name.encode('utf8'))]) + name.encode('utf8') +
                 bytes([len(server.get('host','localhost'))]) + server.get('host','localhost').encode('utf8') +
                 struct.pack("!H", server.get('port', 2355)))
  hosts=zlib.compress(b''.join(hosts))
  data = (bytes([1*key+2*rwd_keys + 4*validate_password + 8*userlist + 16+delete_upgraded, threshold]) +
          mk +
          hosts)

  qr = QrCode.encode_binary(data, QrCode.Ecc.LOW)
  if key:
    clearmem(mk)
    clearmem(data)
  if output=='txt':
    print_qr(qr)
  else:
    print(qr.to_svg_str(2))

def ostore_handler(m, pwd, params, newpwd=None):
    op, keyid, args = ostore.parse(params)
    user = keyid
    host = 'opaque store'

    if op == ostore.changepwd:
        args.insert(0, {'m': Multiplexer,
                        'servers': servers,
                        'change': change,
                        'commit': commit,
                        'user': user,
                        'host': host,
                        'pwd': pwd,
                        'newpwd': newpwd})
    elif op == ostore.erase:
        args.insert(0, {'m': Multiplexer,
                        'servers': servers,
                        'delete': delete,
                        'pwd': pwd,
                        'user': user,
                        'host': host})

    if op == ostore.store:
        rwd = create(m, pwd, user, host)
    else:
        rwd = get(m, pwd, user, host)
    op(rwd, keyid, *args)

    return True

def usage(params, help=False):
  print("usage:")
  print("SPHINX style passwords")
  print("       %s init" % params[0])
  print("       echo -n 'password' | %s <create|change> <user> <site> <[u][l][d][s] [<size>] [<symbols>]> | [<target password>]" % params[0])
  print("       echo -n 'password' | %s get <user> <site>" % params[0])
  if rwd_keys: print("       echo -n 'password' | %s <commit|undo|delete> <user> <site>" % params[0])
  else: print("       %s <commit|undo|delete> <user> <site>" % params[0])
  if userlist: print("       %s list <site>" % params[0])
  print("                            %s healthcheck" % params[0])
  print("       %s qr [svg] [key]" % params[0])
  if ostore.available:
      ostore.usage(params)
  if help: sys.exit(0)
  sys.exit(100)

def arg_rules(params):
  user = params[2]
  site = params[3]
  size = None
  symbols = None
  classes = None
  target = None
  for param in params[4:]:
    if not classes and set(list(param)) - {'u','l','s','d'} == set():
      if 's' in param:
        symbols = bin2pass.symbols
        classes = ''.join(set(param) - set(['s']))
      else:
        classes = param
        symbols = ''
      continue
    if not size:
      try:
        tmp = int(param)
        if tmp<79:
          size = tmp
          continue
      except: pass
    if set(param) - set(bin2pass.symbols) == set():
      symbols = param
      continue
    if verbose: print(f'using "{param}" as target password', file=sys.stderr)
    target = param
  if target is not None and (symbols or classes or size):
    print(f"invalid args for {param[1]}: \"{params[4:]}\"", file=sys.stderr)
    usage(param)
  return user, site, classes or 'uld', symbols if symbols is not None else bin2pass.symbols, size or 0, target

def test_pwd(pwd):
  if zxcvbn is None: return
  q = zxcvbn(pwd.decode('utf8'))
  print("your %s%s (%s/4) master password can be online recovered in %s, and offline in %s, trying ~%s guesses" %
        ("★" * q['score'],
         "☆" * (4-q['score']),
         q['score'],
         q['crack_times_display']['online_throttling_100_per_hour'],
         q['crack_times_display']['offline_slow_hashing_1e4_per_second'],
         q['guesses']), file=sys.stderr)

#### main ####

def main(params=sys.argv):
  #import ctypes
  #libc = ctypes.cdll.LoadLibrary('libc.so.6')
  #fdopen = libc.fdopen
  #log_file = ctypes.c_void_p.in_dll(pyoprf.liboprf,'log_file')
  #fdopen.restype = ctypes.c_void_p
  #log_file.value = fdopen(2, 'w')

  if len(params) < 2: usage(params, True)
  m = Multiplexer(servers)
  cmd = None
  args = []
  if params[1] in ('help', '-h', '--help'):
    usage(params, True)
  elif params[1] == 'create':
    try:
      user,site,classes, syms, size, target = arg_rules(params)
    except: usage(params)
    cmd = create
    args = (user, site, classes, syms, size, target)
  elif params[1] == 'init':
    if len(params) != 2: usage(params)
    sys.exit(init())
  elif params[1] == 'healthcheck':
    cmd = healthcheck
    args = tuple()
  elif params[1] == 'get':
    if len(params) != 4: usage(params)
    cmd = get
    args = (params[2], params[3])
  elif params[1] == 'change':
    try:
      user,site,classes,syms,size, target = arg_rules(params)
    except: usage(params)
    cmd = change
    args = (user, site, classes, syms, size, target)
  elif params[1] == 'commit':
    if len(params) != 4: usage(params)
    cmd = commit
    args = (params[2], params[3])
  elif params[1] == 'delete':
    if len(params) != 4: usage(params)
    cmd = delete
    args = (params[2], params[3])
  elif userlist and params[1] == 'list':
    if len(params) != 3: usage(params)
    cmd = users
    args = (params[2],)
  elif params[1] == 'undo':
    if len(params) != 4: usage(params)
    cmd = undo
    args = (params[2],params[3])
  elif params[1] == 'qr':
    cmd = qrcode
    output = 'txt'
    key = False
    if "svg" in params:
      output="svg"
      del params[params.index("svg")]
    if "key" in params:
      key=True
      del params[params.index("key")]
    if params[2:]: usage(params)
    qrcode(output, key)
    return
  elif ostore.available and ostore.is_cmd(params):
    cmd = ostore_handler
    args = [params]
  else:
    usage(params)

  error = None
  if cmd not in {users, healthcheck}:
    pwd = ''
    if (rwd_keys or cmd in {create,change,get, ostore_handler}):
      pwd = getpwd()
      if cmd == change or (cmd == ostore_handler and params[1] == 'changepwd'):
        newpwd = getpwd()
        if not newpwd:
          newpwd = pwd
        test_pwd(newpwd)
        if cmd == change:
            args=(newpwd,) + args
        else:
            args=args + [newpwd]
      if cmd == create:
        test_pwd(pwd)
    try:
      m.connect()
      ret = cmd(m, pwd, *args)
    except Exception as exc:
      error = exc
      ret = False
      if debug: raise # only for dbg
    clearmem(pwd)
  else:
    try:
      m.connect()
      ret = cmd(m,  *args)
    except Exception as exc:
      error = exc
      ret = False
      if debug: raise # only for dbg
  m.close()

  if not ret:
    if not error:
        print("fail", file=sys.stderr)
        sys.exit(3) # error not handled by exception
    print(error, file=sys.stderr)
    if str(error) == "ERROR: bad checkdigit":
      sys.exit(2) # bad check digit
    sys.exit(1) # generic errors

  if cmd not in {delete, undo, commit, ostore_handler, healthcheck}:
    print(ret)
    sys.stdout.flush()
    clearmem(ret)
  elif ret != True:
    print("reached code that should not be reachable: ", ret)

if __name__ == '__main__':
  try:
    main(sys.argv)
  except Exception:
    print("fail", file=sys.stderr)
    if debug: raise # only for dbg



#!/usr/bin/env python

# SPDX-FileCopyrightText: 2018, Marsiske Stefan 
# SPDX-License-Identifier: GPL-3.0-or-later

import os
#from distutils.core import setup, Extension
from setuptools import setup


# Utility function to read the README file.
# Used for the long_description.  It's nice, because now 1) we have a top level
# README file and 2) it's easier to type in the README file than to put a raw
# string in below ...
def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

from setuptools.command.sdist import sdist as SetuptoolsSdist
class BuildMakefilesSdist(SetuptoolsSdist):
    def run(self):
        os.chdir('man')
        os.system('make')
        os.chdir('..')
        SetuptoolsSdist.run(self)
from setuptools.command.build import build as SetuptoolsBuild
class BuildMakefilesBuild(SetuptoolsBuild):
    def run(self):
        os.chdir('man')
        os.system('make')
        os.chdir('..')
        SetuptoolsBuild.run(self)

setup(name = 'pwdsphinx',
       version = '2.0.3',
       description = 'SPHINX password protocol',
       license = "GPLv3",
       author = 'Stefan Marsiske',
       author_email = 'sphinx@ctrlc.hu',
       url = 'https://github.com/stef/pwdsphinx',
       long_description=read('README.md'),
       long_description_content_type="text/markdown",
       packages = ['pwdsphinx', 'pwdsphinx.converters'],
       install_requires = ("pysodium", "SecureString",
                           "qrcodegen","zxcvbn-python", 'pyequihash', 'pyoprf >= 0.6.0'),
       classifiers = ["Development Status :: 4 - Beta",
                      "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
                      "Topic :: Security :: Cryptography",
                      "Topic :: Security",
                   ],
       entry_points = {
           'console_scripts': [
               'oracle = pwdsphinx.oracle:main',
               'sphinx = pwdsphinx.sphinx:main',
               'websphinx = pwdsphinx.websphinx:main',
               'bin2pass = pwdsphinx.bin2pass:main',
               'sphage = pwdsphinx.converters.sphage:main',
           ],
       },
       cmdclass={'sdist': BuildMakefilesSdist,
                 'build': BuildMakefilesBuild},
       #ext_modules = [libsphinx],
)
